# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""An NLP solver."""

import casadi as ca
import numpy as np
import logging
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import GlobalSettings, Settings
from camino.utils import colored
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class NlpSolver(SolverClass):
    """
    Create NLP solver.

    This solver solves an NLP problem. This is either relaxed or
    the binaries are fixed.
    """

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        """Create NLP problem."""
        super(NlpSolver, self).__init__(problem, stats, s)
        options = regularize_options(
            s.IPOPT_SETTINGS,
            {
                "calc_multipliers": True,
                "ipopt.expect_infeasible_problem": "yes",
                "error_on_fail": False,
            },
            s,
        )

        self.idx_x_integer = problem.idx_x_integer
        # self.callback = DebugCallBack(
        #     'mycallback', problem.x.shape[0],
        #     problem.g.shape[0], problem.p.shape[0]
        # )
        # self.callback.add_to_solver_opts(options, 50)
        if problem.idx_g_dwelltime is not None:
            new_g_constraint = problem.g[problem.idx_g_without_dwelltime]
        else:
            new_g_constraint = problem.g
        self.idx_g_without_dwelltime = problem.idx_g_without_dwelltime
        self.g_fn = ca.Function("g", [problem.x, problem.p], [new_g_constraint])

        if problem.precompiled_nlp is not None:
            self.solver = ca.nlpsol("nlp", "ipopt", problem.precompiled_nlp, options)
        else:
            options.update(
                {
                    "jit": s.WITH_JIT,
                }
            )
            self.solver = ca.nlpsol(
                "nlpsol",
                "ipopt",
                {"f": problem.f, "g": new_g_constraint, "x": problem.x, "p": problem.p},
                options,
            )

    def solve(self, nlpdata: MinlpData, integers_relaxed=True) -> MinlpData:
        """Solve NLP."""
        success_out = []
        sols_out = []

        for sol in nlpdata.solutions_all:
            lbx = nlpdata.lbx
            ubx = nlpdata.ubx
            if not integers_relaxed:
                # Remove integer errors
                x_bin_var = np.round(to_0d(sol["x"][self.idx_x_integer]))
                lbx[self.idx_x_integer] = x_bin_var
                ubx[self.idx_x_integer] = x_bin_var

            sol_new = self.solver(
                p=nlpdata.p,
                x0=nlpdata.x0,
                lbx=lbx,
                ubx=ubx,
                lbg=nlpdata.lbg[self.idx_g_without_dwelltime].flatten(),
                ubg=nlpdata.ubg[self.idx_g_without_dwelltime].flatten(),
            )

            success, stats = self.collect_stats("NLP", sol=sol_new)
            if not success:
                return_status_ok = stats["return_status"] in [
                    "Search_Direction_Becomes_Too_Small",
                    "Maximum_Iterations_Exceeded",
                    "Maximum_CpuTime_Exceeded",
                    "Maximum_WallTime_Exceeded",
                    "Solved_To_Acceptable_Level",
                    "Feasible_Point_Found",
                    "Not_Enough_Degrees_Of_Freedom",
                    "Insufficient_Memory",
                ]
                if return_status_ok:
                    gk = self.g_fn(sol_new["x"], nlpdata.p).full()
                    if np.all(gk <= nlpdata.ubg) and np.all(gk >= nlpdata.lbg):
                        success = True
                    # if np.all(gk <= nlpdata.ubg + self.settings.CONSTRAINT_TOL) and \
                    #         np.all(gk >= nlpdata.lbg - self.settings.CONSTRAINT_TOL):
                    #     success = True
            if not success:
                logger.warning(colored("NLP not solved.", "yellow"))

            success_out.append(success)
            sols_out.append(sol_new)

        nlpdata.prev_solutions = sols_out
        nlpdata.solved_all = success_out

        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Example on how to manually specify a problem from the MINLPlib"""

from camino.problems.dsc import Description
import casadi as ca


def alan():
    dsc = Description()
    x1 = dsc.sym("x1", 1, 0, ca.inf, 0.302884615384618)
    x2 = dsc.sym("x2", 1, 0, ca.inf, 0.0865384615384593)
    x3 = dsc.sym("x3", 1, 0, ca.inf, 0.504807692307693)
    x4 = dsc.sym("x4", 1, 0, ca.inf, 0.10576923076923)
    b5 = dsc.sym_bool("b5", 1, 0)
    b6 = dsc.sym_bool("b6", 1, 0)
    b7 = dsc.sym_bool("b7", 1, 0)
    b8 = dsc.sym_bool("b8", 1, 0)

    dsc.f = (
        x1 * (4 * x1 + 3 * x2 - x3)
        + x2 * (3 * x1 + 6 * x2 + x3)
        + x3 * (-x1 + x2 + 10 * x3)
    )
    dsc.eq(x1 + x2 + x3 + x4, 1)
    dsc.eq(8 * x1 + 9 * x2 + 12 * x3 + 7 * x4, 10)
    dsc.leq(x1 - b5, 0)
    dsc.leq(x2 - b6, 0)
    dsc.leq(x3 - b7, 0)
    dsc.leq(x4 - b8, 0)
    dsc.leq(b5 + b6 + b7 + b8, 3)
    return dsc.get_problem(), dsc.get_data()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Test solver utilities."""

import unittest
import casadi as ca
from camino.settings import GlobalSettings
from camino.solvers import (
    MinlpProblem,
    get_idx_linear_bounds_binary_x,
    get_idx_linear_bounds,
)


class TestSolverUtils(unittest.TestCase):
    """Test utilities for the solvers."""

    def test_get_idx_linear_bounds(self):
        """Get indices for linear bounds."""
        x = GlobalSettings.CASADI_VAR.sym("x", 4)
        f = x[0] + x[1] + x[2] + x[3]
        g = [
            x[0] + x[2],  # Linear on binary variables
            x[0] ** 2 + x[2],  # Non-linear on binary
            x[1] ** 2 + x[3],  # Non-linear
            x[0] - x[2],  # Linear on binary
            x[1] + x[3],  # Linear
        ]
        p = GlobalSettings.CASADI_VAR.sym("p", 0)

        problem = MinlpProblem(f=f, g=ca.vcat(g), x=x, p=p, idx_x_integer=[0, 2])
        idx_lin_bin = get_idx_linear_bounds_binary_x(problem)
        self.assertEqual(idx_lin_bin.tolist(), [0, 3])
        idx_lin = get_idx_linear_bounds(problem)
        self.assertEqual(idx_lin.tolist(), [0, 3, 4])


if __name__ == "__main__":
    unittest.main()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass, field
from typing import List, Dict, Any, Optional
import casadi as ca
from copy import deepcopy


@dataclass
class MinlpData:
    """
    Collection of the data defining a MINLP.

    It is required to fill in:
    - p: list of parameters of the problem
    - x0: initial guess of the problem
    - _lbx: lower bound on the optimization variable x
    - _ubx: upper bound on the optimization variable x
    - _lbg: lower bound on the constraint vector g
    - _ubg: upper bound on the constraint vector g
    """

    p: List[float]
    x0: ca.DM
    _lbx: ca.DM
    _ubx: ca.DM
    _lbg: ca.DM
    _ubg: ca.DM
    _prev_solutions: Optional[List[Dict[str, Any]]] = None
    solved_all: List[bool] = field(default_factory=lambda: [True])
    best_solutions: Optional[list] = field(default_factory=lambda: [])
    relaxed: bool = False

    @property
    def nr_sols(self):
        if self._prev_solutions:
            return len(self._prev_solutions)
        else:
            return 1

    @property
    def solved(self):
        return self.solved_all[0]

    @solved.setter
    def solved(self, value):
        self.solved_all = [value]

    @property
    def _sol(self):
        """Get safely previous solution."""
        if self._prev_solutions is not None:
            return self._prev_solutions[0]
        else:
            return {"f": -ca.inf, "x": ca.DM(self.x0)}

    @property
    def solutions_all(self):
        if self._prev_solutions is not None:
            return self._prev_solutions
        else:
            return [{"f": -ca.inf, "x": ca.DM(self.x0)}]

    @property
    def prev_solution(self):
        """Previous solution."""
        raise Exception("May not be used!")

    @property
    def obj_val(self):
        """Get float value."""
        return float(self._sol["f"])

    @property
    def x_sol(self):
        """Get x solution."""
        return self._sol["x"]

    @property
    def lam_g_sol(self):
        """Get lambda g solution."""
        return self._sol["lam_g"]

    @property
    def lam_x_sol(self):
        """Get lambda g solution."""
        return self._sol["lam_x"]

    @prev_solution.setter
    def prev_solution(self, value):
        self.prev_solutions = [value]
        self.relaxed = False

    @property
    def prev_solutions(self):
        """Get all previous solutions."""
        return self._prev_solutions

    @prev_solutions.setter
    def prev_solutions(self, value):
        self._prev_solutions = value
        self.relaxed = False

    @property
    def lbx(self):
        """Get lbx."""
        return deepcopy(self._lbx)

    @property
    def ubx(self):
        """Get ubx."""
        return deepcopy(self._ubx)

    @property
    def lbg(self):
        """Get lbx."""
        return deepcopy(self._lbg)

    @property
    def ubg(self):
        """Get ubx."""
        return deepcopy(self._ubg)

    @lbg.setter
    def lbg(self, value):
        self._lbg = value

    @ubg.setter
    def ubg(self, value):
        self._ubg = value

    @lbx.setter
    def lbx(self, value):
        self._lbx = value

    @ubx.setter
    def ubx(self, value):
        self._ubx = value



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Bonmin solver."""

import casadi as ca
from camino.solvers import (
    MiSolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import Settings
from camino.utils import tic, toc


class BonminSolver(MiSolverClass):
    """Create MINLP solver (using bonmin)."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        s: Settings,
        algo_type="B-BB",
    ):
        """Create MINLP problem.

        :param algo_type: Algorithm type, options: B-BB, B-OA, B-QG, or B-Hyb
        """
        super(BonminSolver, self).__init__(problem, data, stats, s)
        options = regularize_options(s.BONMIN_SETTINGS, {}, s)

        self.nr_x = problem.x.shape[0]
        discrete = [0] * self.nr_x
        for i in problem.idx_x_integer:
            discrete[i] = 1
        options.update(
            {
                "discrete": discrete,
                "bonmin.algorithm": algo_type,
            }
        )
        minlp = {"f": problem.f, "g": problem.g, "x": problem.x, "p": problem.p}
        self.solver = ca.nlpsol("minlp", "bonmin", minlp, options)

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve MINLP."""
        tic()
        nlpdata.prev_solution = self.solver(
            x0=nlpdata.x0,
            lbx=nlpdata.lbx,
            ubx=nlpdata.ubx,
            lbg=nlpdata.lbg,
            ubg=nlpdata.ubg,
            p=nlpdata.p,
        )
        nlpdata.solved, stats = self.collect_stats(
            "MINLP", sol=nlpdata.prev_solutions[-1]
        )
        self.stats["total_time_calc"] = toc(reset=True)
        return nlpdata

    def reset(self, nlpdata: MinlpData):
        """Reset problem data."""

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart the algorithm."""



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Main running scripts."""

import casadi as ca
from camino.solver import MinlpSolver
from camino.problem import MetaDataOcp
from camino.settings import Settings
from camino.problems.problem_collection import PROBLEMS
from camino.utils import make_bounded, logger, colored
from camino.utils.data import write_json, read_json
from camino.utils.validate import check_solution


def batch_runner(algorithm, target, nl_files):
    """Run a batch of problems."""
    from os import makedirs, path
    from time import time

    def do_write(overview_target, start, i, algorithm, total_stats):
        time_now = time() - start
        total_time = time_now / (i + 1) * total_to_compute
        write_json(
            {
                "time": time_now,
                "total": total_to_compute,
                "done": (i + 1),
                "progress": (i + 1) / total_to_compute,
                "time_remaining_est": total_time - time_now,
                "time_total_est": total_time,
                "algorithm": algorithm,
                "data": total_stats,
            },
            overview_target,
        )

    overview_target = path.join(target, "overview.json")
    start = time()
    total_to_compute = len(nl_files)
    if path.exists(overview_target):
        data = read_json(overview_target)
        total_stats = data["data"]
        algorithm = data["algorithm"]
        i_start = data["done"]
        start -= data["time"]
        if total_stats[-1][0] != i_start:
            total_stats.append([i_start, nl_files[i_start], -ca.inf, "FAILED", "CRASH"])
        do_write(overview_target, start, i_start, algorithm, total_stats)
        i_start += 1
    else:
        makedirs(target, exist_ok=True)
        total_stats = [
            [
                "id",
                "path",
                "obj",
                "dual_obj",
                "load_time",
                "calc_time",
                "solver_time",
                "python_time",
                "iter_nr",
                "NLP_runs",
                "FNLP_runs",
                "MIQP_runs",
                "MILP_runs",
            ]
        ]
        i_start = 0

    for i in range(i_start, len(nl_files)):
        nl_file = nl_files[i]
        try:
            stats, data = runner(algorithm, "nl_file", None, [nl_file])
            stats["x_star"] = data.x_sol
            stats["f_star"] = data.obj_val
            stats.print()
            stats.save(path.join(target, f"stats_{i}.pkl"))
            total_stats.append(
                [
                    i,
                    nl_file,
                    data.obj_val,
                    stats.data["lb"],
                    stats["total_time_loading"],
                    stats["total_time_calc"],
                    stats["t_solver_total"],
                    stats["t_python_solver"],
                    stats["iter_nr"],
                    stats["NLP.runs"],
                    stats["FC-NLP.runs"],
                    stats["BR-MIQP.runs"],
                    stats["LB-MILP.runs"],
                ]
            )
        except Exception as e:
            print(f"{e}")
            total_stats.append([i, nl_file, -ca.inf, "FAILED", f"{e}"])
        do_write(overview_target, start, i, algorithm, total_stats)


def runner(solver_name, problem_name, target_file, args):
    """General runner."""
    if problem_name not in PROBLEMS:
        raise Exception(f"No {problem_name=}, available: {PROBLEMS.keys()}")

    if args is None:
        args = []

    colored(f"Load problem {problem_name} with args: {args}", color="green")

    output = PROBLEMS[problem_name](*args)
    if len(output) == 2:
        problem, data = output
        settings = Settings()
        logger.info("Using default settings")
    else:
        logger.info("Using custom settings")
        problem, data, settings = output

    make_bounded(problem, data, new_inf=1e3)

    if len(problem.idx_x_integer) == 0:
        solver_name = "relaxed"

    solver = MinlpSolver(
        solver_name, problem, data, settings=settings, problem_name=problem_name
    )

    logger.info(f"Start mode {solver_name}")
    data = solver.solve(data)
    solved, stats = solver.collect_stats()
    stats.print()

    if target_file is not None:
        if (
            isinstance(problem.meta, MetaDataOcp)
            and problem.meta.dump_solution is not None
        ):
            output_data = problem.meta.dump_solution(data.x_sol)
        else:
            output_data = data.x_sol

        write_json({"w0": output_data}, target_file)

    logger.info(f"Objective value: {data.obj_val}")
    check_solution(problem, data, data.x_sol, settings)
    return stats, data



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Set of solvers based on outer approximation."""

import logging
import casadi as ca
import numpy as np
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    get_idx_linear_bounds,
    regularize_options,
    get_idx_inverse,
    extract_bounds,
)
from camino.settings import GlobalSettings, Settings

logger = logging.getLogger(__name__)


class OuterApproxMILP(SolverClass):
    r"""
    Outer approximation.

    This implementation assumes the following input:
        min f(x)
        s.t. lb < g(x)

    It constructs the following problem:
        min \alpha
        s.t.
            \alpha \geq f(x) + \nabla (x) (x-x^i)
            lb \leq g(x) + \nabla  g(x) (x-x^i)
    """

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        s: Settings,
        with_lin_bounds=True,
    ):
        """Create the outer approximation master problem."""
        super(OuterApproxMILP, self).__init__(problem, stats, s)
        self.setup_common(problem, s)
        # Last variable is alpha for epigraph reformulation
        self._x = GlobalSettings.CASADI_VAR.sym("x", self.nr_x)
        self._alpha = GlobalSettings.CASADI_VAR.sym("alpha")
        self.problem = problem
        OuterApproxMILP.reset(self, data)

    def setup_common(self, problem: MinlpProblem, s: Settings):
        """Set up common data"""
        self.settings = s
        self.options = regularize_options(s.MIP_SETTINGS, {}, s)
        self.f = ca.Function(
            "f", [problem.x, problem.p], [problem.f], {"jit": s.WITH_JIT}
        )
        self.grad_f = ca.Function(
            "gradient_f_x",
            [problem.x, problem.p],
            [ca.gradient(problem.f, problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.g = ca.Function(
            "g", [problem.x, problem.p], [problem.g], {"jit": s.WITH_JIT}
        )
        self.jac_g = ca.Function(
            "gradient_g_x",
            [problem.x, problem.p],
            [ca.jacobian(problem.g, problem.x)],
            {"jit": s.WITH_JIT},
        )

        self.nr_x = problem.x.shape[0]
        self.nr_g_orig = problem.g.shape[0]
        self.idx_x_integer = problem.idx_x_integer
        self.nr_x_bin = len(problem.idx_x_integer)
        self.options["discrete"] = [
            1 if i in self.idx_x_integer else 0 for i in range(self.nr_x)
        ] + [0]

    def add_solution(self, nlpdata, solved, solution, integers_relaxed=False):
        """Add a cut."""
        x_sol = solution["x"][: self.nr_x]
        if solved and (not integers_relaxed):
            self._g = ca.vertcat(
                self._g,
                self.f(x_sol, nlpdata.p)
                + self.grad_f(x_sol, nlpdata.p).T @ (self._x - x_sol)
                - self._alpha,
            )
            self._lbg.append(-ca.inf)
            self._ubg.append(0)

        g_lin = self.g(x_sol, nlpdata.p)
        jac_g = self.jac_g(x_sol, nlpdata.p)
        self._g = ca.vertcat(self._g, g_lin + jac_g @ (self._x - x_sol))
        self._lbg.append(nlpdata.lbg)
        self._ubg.append(nlpdata.ubg)

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """Solve the outer approximation master problem (MILP)."""
        for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
            self.add_solution(nlpdata, solved, solution, integers_relaxed)

        x_sol = nlpdata.x_sol[: self.nr_x]

        solver = ca.qpsol(
            f"oa_with_{self._g.shape[0]}_cut",
            self.settings.MIP_SOLVER,
            {
                "f": self._alpha,
                "g": self._g,
                "x": ca.vertcat(self._x, self._alpha),
            },
            self.options,
        )

        solution = solver(
            x0=ca.vertcat(x_sol, nlpdata.obj_val),
            lbx=ca.vertcat(nlpdata.lbx, -1e8),
            ubx=ca.vertcat(nlpdata.ubx, ca.inf),
            lbg=ca.vertcat(*self._lbg),
            ubg=ca.vertcat(*self._ubg),
        )
        x_full = solution["x"].full()[: self.nr_x]
        solution["x"] = x_full
        nlpdata.prev_solution = solution
        nlpdata.solved, stats = self.collect_stats("OA-MILP", solver, solution)
        return nlpdata

    def reset(self, nlpdata: MinlpData):
        """Reset this problem."""
        self._g = np.array([])
        self._lbg = []
        self._ubg = []

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart cuts."""
        relaxed = self.stats.relaxed_solution
        if self.stats.relaxed:
            self.add_solution(nlpdata, True, relaxed.solutions_all[0], True)

        if not nlpdata.relaxed:
            for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
                self.add_solution(nlpdata, solved, solution)


class OuterApproxMIQP(OuterApproxMILP):
    "Implementation of quadratic outer approximation"

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        s: Settings,
        hess: str = "f",
    ):
        """Create the quadratic outer approximation master problem."""
        super(OuterApproxMIQP, self).__init__(problem, data, stats, s)
        self.hessian_type = hess
        if self.hessian_type == "f":
            self.hess = ca.Function(
                "hess_f",
                [problem.x, problem.p],
                [ca.hessian(problem.f, problem.x)[0]],
                {"jit": self.settings.WITH_JIT},
            )
        elif self.hessian_type == "lag":
            lam = GlobalSettings.CASADI_VAR.sym("lam", self.nr_g_orig)
            lagrangian = problem.f + lam.T @ problem.g
            lagrangian_hess = ca.hessian(lagrangian, ca.vertcat(problem.x, lam))[0][
                : self.nr_x, : self.nr_x
            ]
            self.hess = ca.Function(
                "hess_lag",
                [problem.x, lam, problem.p],
                [lagrangian_hess],
                {"jit": self.settings.WITH_JIT},
            )
        else:
            raise AttributeError(
                "Available Hessian types: 'f' for standard objective Hessian, 'lag' for Hessian of the Lagrangian"
            )

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """Solve the quadratic outer approximation master problem (MIQP)."""
        for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
            self.add_solution(nlpdata, solved, solution, integers_relaxed)

        x_sol = nlpdata.x_sol[: self.nr_x]
        if self.hessian_type == "f":
            hess = self.hess(x_sol, nlpdata.p)
        elif self.hessian_type == "lag":
            hess = self.hess(x_sol, nlpdata.lam_g_sol, nlpdata.p)
        f_quad = self._alpha + 0.5 * (self._x - x_sol).T @ hess @ (self._x - x_sol)

        solver = ca.qpsol(
            f"oa_with_{self._g.shape[0]}_cut",
            self.settings.MIP_SOLVER,
            {
                "f": f_quad,
                "g": self._g,
                "x": ca.vertcat(self._x, self._alpha),
            },
            self.options,
        )

        solution = solver(
            x0=ca.vertcat(x_sol, nlpdata.obj_val),
            lbx=ca.vertcat(nlpdata.lbx, -1e8),
            ubx=ca.vertcat(nlpdata.ubx, ca.inf),
            lbg=ca.vertcat(*self._lbg),
            ubg=ca.vertcat(*self._ubg),
        )
        x_full = solution["x"].full()[: self.nr_x]
        solution["x"] = x_full
        nlpdata.prev_solution = solution
        nlpdata.solved, stats = self.collect_stats("OA-MIQP", solver, solution)
        return nlpdata


class OuterApproxMILPImproved(OuterApproxMILP):
    """i
    Improved version of outer approximation.

    Cutting planes are added only for linear constraints to avoid the creation of infeasible master problems.
    """

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create the improved outer approximation master problem."""
        super(OuterApproxMILPImproved, self).__init__(problem, data, stats, s)

        self.idx_g_lin = get_idx_linear_bounds(problem)
        self.idx_g_nonlin = get_idx_inverse(self.idx_g_lin, problem.g.shape[0])
        self.g_lin = ca.Function(
            "g",
            [problem.x, problem.p],
            [problem.g[self.idx_g_lin]],
            {"jit": s.WITH_JIT},
        )
        self.g_nonlin = ca.Function(
            "g",
            [problem.x, problem.p],
            [problem.g[self.idx_g_nonlin]],
            {"jit": s.WITH_JIT},
        )
        self.jac_g_nonlin = ca.Function(
            "gradient_g_x",
            [problem.x, problem.p],
            [ca.jacobian(problem.g[self.idx_g_nonlin], problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.reset(data)

    def reset(self, data):
        """Reset internal data."""
        if self.idx_g_lin.shape[0] > 0:
            _, self._g, self._lbg, self._ubg = extract_bounds(
                self.problem, data, self.idx_g_lin, self._x, self.problem.idx_x_integer
            )
        else:
            _, self._g, self._lbg, self._ubg = 0, [], [], []

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """Solve the improved outer approximation master problem (MILP)."""
        for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
            self.add_solution(nlpdata, solved, solution, integers_relaxed)

        x_sol = nlpdata.x_sol[: self.nr_x]

        solver = ca.qpsol(
            f"oa_with_{self._g.shape[0]}_cut",
            self.settings.MIP_SOLVER,
            {
                "f": self._alpha,
                "g": self._g,
                "x": ca.vertcat(self._x, self._alpha),
            },
            self.options,
        )

        solution = solver(
            x0=ca.vertcat(x_sol, nlpdata.obj_val),
            lbx=ca.vertcat(nlpdata.lbx, -1e8),
            ubx=ca.vertcat(nlpdata.ubx, ca.inf),
            lbg=ca.vertcat(*self._lbg),
            ubg=ca.vertcat(*self._ubg),
        )
        x_full = solution["x"].full()[: self.nr_x]
        solution["x"] = x_full
        nlpdata.prev_solution = solution
        nlpdata.solved, stats = self.collect_stats("OAI-MILP", solver, solution)
        return nlpdata


class OuterApproxMIQPImproved(OuterApproxMILP):
    """
    Improved version of quadratic outer approximation.

    Cutting planes are added only for linear constraints to avoid the creation of infeasible master problems.
    """

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        s: Settings,
        hess: str = "f",
    ):
        """Create improved version of the quadratic outer approximation master problem."""
        super(OuterApproxMIQPImproved, self).__init__(problem, data, stats, s)

        self.hessian_type = hess
        if self.hessian_type == "f":
            self.hess = ca.Function(
                "hess_f",
                [problem.x, problem.p],
                [ca.hessian(problem.f, problem.x)[0]],
                {"jit": self.settings.WITH_JIT},
            )
        elif self.hessian_type == "lag":
            lam = GlobalSettings.CASADI_VAR.sym("lam", self.nr_g_orig)
            lagrangian = problem.f + lam.T @ problem.g
            lagrangian_hess = ca.hessian(lagrangian, ca.vertcat(problem.x, lam))[0][
                : self.nr_x, : self.nr_x
            ]
            self.hess = ca.Function(
                "hess_lag",
                [problem.x, lam, problem.p],
                [lagrangian_hess],
                {"jit": self.settings.WITH_JIT},
            )
        else:
            raise AttributeError(
                "Available Hessian types: 'f' for standard objective Hessian, 'lag' for Hessian of the Lagrangian"
            )

        self.idx_g_lin = get_idx_linear_bounds(problem)
        self.idx_g_nonlin = get_idx_inverse(self.idx_g_lin, problem.g.shape[0])
        self.g_lin = ca.Function(
            "g",
            [problem.x, problem.p],
            [problem.g[self.idx_g_lin]],
            {"jit": s.WITH_JIT},
        )
        self.g_nonlin = ca.Function(
            "g",
            [problem.x, problem.p],
            [problem.g[self.idx_g_nonlin]],
            {"jit": s.WITH_JIT},
        )
        self.jac_g_nonlin = ca.Function(
            "gradient_g_x",
            [problem.x, problem.p],
            [ca.jacobian(problem.g[self.idx_g_nonlin], problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.reset(data)

    def reset(self, data):
        """Reset internal data."""
        if self.idx_g_lin.shape[0] > 0:
            _, self._g, self._lbg, self._ubg = extract_bounds(
                self.problem, data, self.idx_g_lin, self._x, self.problem.idx_x_integer
            )
        else:
            _, self._g, self._lbg, self._ubg = 0, [], [], []

    def solve(self, nlpdata: MinlpData, integers_relaxed=False) -> MinlpData:
        """Solve the improved quadratic outer approximation master problem (MIQP)."""
        for solved, solution in zip(nlpdata.solved_all, nlpdata.solutions_all):
            self.add_solution(nlpdata, solved, solution, integers_relaxed)

        x_sol = nlpdata.x_sol[: self.nr_x]

        if self.hessian_type == "f":
            hess = self.hess(x_sol, nlpdata.p)
        elif self.hessian_type == "lag":
            hess = self.hess(x_sol, nlpdata.lam_g_sol, nlpdata.p)
        f_quad = self._alpha + 0.5 * (self._x - x_sol).T @ hess @ (self._x - x_sol)

        solver = ca.qpsol(
            f"oa_with_{self._g.shape[0]}_cut",
            self.settings.MIP_SOLVER,
            {
                "f": f_quad,
                "g": self._g,
                "x": ca.vertcat(self._x, self._alpha),
            },
            self.options,
        )

        solution = solver(
            x0=ca.vertcat(x_sol, nlpdata.obj_val),
            lbx=ca.vertcat(nlpdata.lbx, -1e8),
            ubx=ca.vertcat(nlpdata.ubx, ca.inf),
            lbg=ca.vertcat(*self._lbg),
            ubg=ca.vertcat(*self._ubg),
        )
        x_full = solution["x"].full()[: self.nr_x]
        solution["x"] = x_full
        nlpdata.prev_solution = solution
        nlpdata.solved, stats = self.collect_stats("OAI-MIQP", solver, solution)
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Base algorithm for feasibility pumps."""

from copy import deepcopy
from camino.settings import Settings
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.stats import Stats
from camino.data import MinlpData
from camino.problem import MinlpProblem
from camino.solvers.utils import any_equal
from camino.solvers.pumps.utils import (
    integer_error,
    create_rounded_data,
    perturbe_x,
    random_perturbe_x,
)
from camino.utils import toc, logging
from camino.utils.conversion import to_0d
from camino.solvers import MiSolverClass

logger = logging.getLogger(__name__)


class PumpBase(MiSolverClass):

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        pump,
        nlp=None,
    ):
        """Create a solver class."""
        super(PumpBase, self).__init__(problem, data, stats, settings)
        self.pump = pump
        self.idx_x_integer = problem.idx_x_integer
        if nlp is None:
            nlp = NlpSolver(problem, stats, settings)
        self.nlp = nlp

    def solve(self, nlpdata: MinlpData, integers_relaxed: bool = False) -> MinlpData:
        """Solve the problem."""
        if self.stats.relaxed_solution is None:
            if not integers_relaxed:
                nlpdata = self.nlp.solve(nlpdata)
            self.stats.relaxed_solution = nlpdata
        else:
            nlpdata = self.stats.relaxed_solution

        relaxed_value = nlpdata.obj_val
        prev_x = []
        distances = [integer_error(nlpdata.x_sol[self.idx_x_integer])]
        while (
            distances[-1] > self.settings.CONSTRAINT_TOL
            and self.stats["iter_nr"] < self.settings.PUMP_MAX_ITER
        ):
            datarounded = create_rounded_data(nlpdata, self.idx_x_integer)
            require_restart = False
            for i, sol in enumerate(datarounded.solutions_all):
                new_x = to_0d(sol["x"])
                perturbe_remaining = self.settings.PARALLEL_SOLUTIONS
                while (
                    any_equal(new_x, prev_x, self.idx_x_integer)
                    and perturbe_remaining > 0
                ):
                    new_x = perturbe_x(
                        to_0d(nlpdata.solutions_all[i]["x"]), self.idx_x_integer
                    )
                    perturbe_remaining -= 1

                datarounded.prev_solutions[i]["x"] = new_x
                if perturbe_remaining == 0:
                    require_restart = True

                prev_x.append(new_x)

            if not require_restart:
                data = self.pump.solve(
                    datarounded, int_error=distances[-1], obj_val=relaxed_value
                )
                distances.append(integer_error(data.x_sol[self.idx_x_integer]))

            if (
                len(distances) > self.settings.PUMP_MAX_STEP_IMPROVEMENTS
                and distances[-self.settings.PUMP_MAX_STEP_IMPROVEMENTS - 1]
                < distances[-1]
            ) or require_restart:
                data.prev_solutions[0]["x"] = random_perturbe_x(
                    data.x_sol, self.idx_x_integer
                )
                data = self.pump.solve(
                    data, int_error=distances[-1], obj_val=relaxed_value
                )
                distances.append(integer_error(data.x_sol[self.idx_x_integer]))

            # Added heuristic, not present in the original implementation
            if distances[-1] < self.settings.CONSTRAINT_INT_TOL:
                datarounded = self.nlp.solve(
                    create_rounded_data(data, self.idx_x_integer), integers_relaxed=False
                )
                if self.update_best_solutions(datarounded):
                    return self.get_best_solutions(datarounded)

            self.stats["iter_nr"] += 1
            logger.info(f"Iteration {self.stats['iter_nr']} finished")

        self.stats["total_time_calc"] += toc(reset=True)
        data = self.nlp.solve(data, integers_relaxed=False)
        self.update_best_solutions(data)
        return self.get_best_solutions(data)

    def reset(self, nlpdata: MinlpData):
        """Reset problem data."""

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart."""
        if not nlpdata.relaxed:
            self.update_best_solutions(nlpdata)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Data utilities for loading and writing data."""

import json
import pickle


def write_json(data, file):
    """Write json file."""
    with open(file, "w") as f:
        json.dump(data, f, indent=4)


def read_json(file):
    """Load json file."""
    with open(file, "r") as f:
        data = json.load(f)
    return data


def load_pickle(file):
    """Load a pickle file."""
    with open(file, "rb") as f:
        return pickle.load(f)


def save_pickle(data, file):
    """Write a pickle file."""
    with open(file, "wb") as handle:
        pickle.dump(data, handle)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""An FNLP solver searching for the closest feasible point."""

import casadi as ca
import numpy as np
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import GlobalSettings, Settings
from camino.utils import colored
from camino.utils.conversion import to_0d
import logging

logger = logging.getLogger(__name__)


class FindClosestNlpSolver(SolverClass):
    """Find closest feasible."""

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        """Create NLP problem."""
        super(FindClosestNlpSolver, self).__init__(problem, stats, s)
        options = regularize_options(s.IPOPT_SETTINGS, {"jit": s.WITH_JIT}, s)
        self.idx_x_integer = problem.idx_x_integer
        x_hat = GlobalSettings.CASADI_VAR.sym("x_hat", len(self.idx_x_integer))
        x_best = GlobalSettings.CASADI_VAR.sym("x_best", len(self.idx_x_integer))
        if problem.idx_g_dwelltime is not None:
            new_g_constraint = problem.g[problem.idx_g_without_dwelltime]
        else:
            new_g_constraint = problem.g
        self.idx_g_without_dwelltime = problem.idx_g_without_dwelltime

        f = ca.norm_2(problem.x[self.idx_x_integer] - x_hat) ** 2
        self.solver = ca.nlpsol(
            "nlpsol",
            "ipopt",
            {
                "f": f,
                "g": ca.vertcat(
                    new_g_constraint,
                    ca.dot(
                        problem.x[self.idx_x_integer] - x_best,
                        problem.x[self.idx_x_integer] - x_best,
                    ),
                ),
                "x": problem.x,
                "p": ca.vertcat(problem.p, x_hat, x_best),
            },
            options,
        )

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve NLP."""
        success_out = []
        sols_out = []
        has_best = len(nlpdata.best_solutions) >= 1
        if has_best:
            x_best = nlpdata.best_solutions[-1]["x"][self.idx_x_integer]
        for success_prev, sol in zip(nlpdata.solved_all, nlpdata.solutions_all):
            if success_prev:
                success_out.append(success_prev)
                sols_out.append(sol)
            else:
                lbx = nlpdata.lbx
                ubx = nlpdata.ubx
                x_bin_var = to_0d(sol["x"][self.idx_x_integer])
                if not has_best:
                    x_best = x_bin_var
                    distance = 1e16
                else:
                    distance = ca.dot(x_best - x_bin_var, x_best - x_bin_var)

                sol_new = self.solver(
                    x0=nlpdata.x0,
                    lbx=lbx,
                    ubx=ubx,
                    lbg=ca.vertcat(
                        nlpdata.lbg[self.idx_g_without_dwelltime].flatten(), 0
                    ),
                    ubg=ca.vertcat(
                        nlpdata.ubg[self.idx_g_without_dwelltime].flatten(), distance
                    ),
                    p=ca.vertcat(nlpdata.p, x_bin_var, x_best),
                )
                sol_new["x_infeasible"] = sol["x"]
                success, _ = self.collect_stats("FC-NLP", sol=sol_new)
                if not success:
                    logger.warning(colored("FC-NLP not solved", "yellow"))

                success_out.append(False)
                sols_out.append(sol_new)

        nlpdata.prev_solutions = sols_out
        nlpdata.solved_all = success_out
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""General problem structure."""

from camino.problem import (
    MinlpProblem,
    MetaData,
    MetaDataMpc,
    MetaDataOcp,
)  # noqa: F401
from camino.data import MinlpData  # noqa: F401



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Set of utilities for MINLP optimization and visualization."""

from math import sqrt
from typing import Union
from matplotlib import pyplot as plt
import matplotlib
import numpy as np
import casadi as ca
import logging
from time import perf_counter
from camino.problems import MinlpData, MinlpProblem, MetaDataOcp

logger = logging.getLogger(__name__)

try:
    from colored import fg, stylize

    def colored(text, color="red"):
        """Color a text."""
        logger.info(stylize(text, fg(color)))

except Exception:

    def colored(text, color=None):
        """Color a text."""
        logger.info(text)


perf_ti = None


def make_bounded(problem: MinlpProblem, data: MinlpData, new_inf=1e3):
    """Make bounded."""
    lbx, ubx = data.lbx, data.ubx
    lbg, ubg = data.lbg, data.ubg

    # # Move all continuous bounds to g!
    # when constraints are one sided the convention is -inf <= g <= ubg
    g_extra, g_lb, g_ub = [], [], []
    for i in range(data.lbx.shape[0]):
        if i in problem.idx_x_integer:
            if lbx[i] < -new_inf:
                lbx[i] = -new_inf
            if ubx[i] > new_inf:
                ubx[i] = new_inf

    data.lbx, data.ubx = lbx, ubx
    # Update
    problem.g = ca.vertcat(problem.g, ca.vertcat(*g_extra))
    data.lbg = np.concatenate((ca.DM(lbg), ca.DM(g_lb)))
    data.ubg = np.concatenate((ca.DM(ubg), ca.DM(g_ub)))


def setup_logger(level=logging.WARNING):
    """Set up the logging."""
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=level
    )
    logging.getLogger("matplotlib").setLevel(logging.ERROR)
    logging.getLogger("PIL").setLevel(logging.ERROR)


def tic():
    """Tic."""
    global perf_ti
    perf_ti = perf_counter()


def toc(reset=False):
    """Toc."""
    global perf_ti
    if perf_ti is None:
        tic()
    tim = perf_counter()
    dt = tim - perf_ti
    logger.info(f"Elapsed time: {dt} s")
    if reset:
        perf_ti = tim
    return dt


def latexify(fig_width=None, fig_height=None):
    """
    Set up matplotlib's RC params for LaTeX plotting.

    Call this before plotting a figure.

    Parameters
    ----------
    fig_width : float, optional, inches
    fig_height : float,  optional, inches
    columns : {1, 2}
    """
    # code adapted from http://www.scipy.org/Cookbook/Matplotlib/LaTeX_Examples

    # Width and max height in inches for IEEE journals taken from
    # computer.org/cms/Computer.org/Journal%20templates/transactions_art_guide.pdf
    if fig_width is None:
        fig_width = 5  # width in inches

    if fig_height is None:
        golden_mean = (sqrt(5) - 1.0) / 2.0  # Aesthetic ratio
        fig_height = fig_width * golden_mean  # height in inches

    params = {
        # "backend": "ps",
        "text.latex.preamble": r"\usepackage{gensymb} \usepackage{amsmath}",
        "axes.labelsize": 9,  # fontsize for x and y labels (was 10)
        "axes.titlesize": 9,
        "lines.linewidth": 2,
        "legend.fontsize": 9,  # was 10
        "xtick.labelsize": 9,
        "ytick.labelsize": 9,
        "text.usetex": True,
        "figure.figsize": [fig_width, fig_height],
        "font.family": "serif",
    }

    matplotlib.rcParams.update(params)


def plot_trajectory(
    x_star: np.ndarray,
    s_collection: Union[np.ndarray, list],
    a_collection: Union[np.ndarray, list],
    meta: MetaDataOcp,
    title: str,
):
    """Plot a trajectory of an OCP."""
    alpha = 0.4
    if isinstance(a_collection, np.ndarray):
        a_collection = [a_collection]
        alpha = 0.7
    if isinstance(s_collection, np.ndarray):
        s_collection = [s_collection]
    if isinstance(meta.scaling_coeff_control, type(None)):
        meta.scaling_coeff_control = [1 for _ in range(a_collection[0].shape[0])]

    N = a_collection[0].shape[0]
    dt = meta.dt
    if dt is None:
        time_array = np.hstack((np.array([0]), x_star[meta.idx_t]))
    else:
        time_array = np.linspace(0, N * dt, N + 1)
    n_controls = meta.n_continuous_control + meta.n_discrete_control

    latexify()
    fig, axs = plt.subplots(meta.n_state + n_controls, 1, figsize=(4.5, 8), sharex=True)
    for s in range(meta.n_state):
        axs[s].grid()
        for s_traj in s_collection:
            if len(s_traj.shape) == 1:
                s_traj = s_traj[..., np.newaxis]
            axs[s].plot(time_array, s_traj[:, s], "-", alpha=alpha, color="tab:blue")
        # axs[s].axhline(s_max[s], linestyle=":", color="k", alpha=0.7)
        # axs[s].axhline(s_min[s], linestyle=":", color="k", alpha=0.7)
        axs[s].set_ylim(
            0,
        )
        axs[s].set_ylabel(f"$x_{s}$")

    for a in range(n_controls):
        if a >= meta.n_continuous_control:
            # Discrete control
            axs[meta.n_state + a].set_ylim(-1.5, 1.5)
        else:
            axs[meta.n_state + a].set_ylim(-10.5, 10.5)

        for a_traj in a_collection:
            if len(a_traj.shape) == 1:
                a_traj = a_traj[..., np.newaxis]
            axs[meta.n_state + a].step(
                time_array,
                meta.scaling_coeff_control[a] * np.append([a_traj[0, a]], a_traj[:, a]),
                alpha=alpha,
                color="tab:orange",
                marker=".",
                where="post",
            )
        axs[meta.n_state + a].set_ylabel(f"$u_{a}$")
        axs[meta.n_state + a].grid()

    axs[0].set_title(title)
    axs[-1].set_xlabel(r"time [sec]")
    axs[-1].set_xlim(0, time_array[-1])
    # axs[-1].grid()
    # axs[-1].axhline(model.a_max, linestyle=":", color="k", alpha=0.5)
    # axs[-1].axhline(model.a_min, linestyle=":", color="k", alpha=0.5)

    plt.tight_layout()
    return fig, axs



"""
Ambient conditions model.

Adrian Buerger, 2022
Adapted by Wim Van Roy and Andrea Ghezzi, 2023
"""

from dataclasses import dataclass, field
import os
import pandas as pd
import numpy as np
from datetime import datetime, timedelta

from camino.settings import GlobalSettings
import logging


logger = logging.getLogger(__name__)


@dataclass
class Timing:
    simplified: bool = False
    earliest_sunrise_time_UTC: tuple = field(init=False)
    latest_sunset_time_UTC: tuple = field(init=False)
    dt_day: int = field(init=False)
    dt_night: int = field(init=False)
    N_day: int = field(init=False)
    N_night: int = field(init=False)
    N_short_term: int = field(init=False)
    t_f: int = field(init=False)
    N: int = field(init=False)

    def __post_init__(self):
        self.earliest_sunrise_time_UTC = (4, 30)
        self.latest_sunset_time_UTC = (21, 30)

        if self.simplified:
            self.dt_day = 3600
            self.dt_night = 1 * self.dt_day
            self.N_day = 17 * 1
            self.N_night = 7 * 1
            self.N_short_term = 1
        else:
            self.dt_day = 900
            self.dt_night = 2 * self.dt_day
            self.N_day = 17 * 4
            self.N_night = 7 * 2
            self.N_short_term = 2

        self.t_f = self.N_day * self.dt_day + self.N_night * self.dt_night
        self.N = self.N_day + self.N_night + (self.N_short_term - 1)


class Ambient:
    """Ambient information."""

    def __init__(self, timing: Timing):
        """Create ambient information."""
        self.timing = timing
        self._load_forecast_data()
        self._fill_remaining_nan_values()
        self._compute_load_forecast()
        self._manual_tuning_of_forecasts()
        self._set_time_steps()

    def _load_forecast_data(self, method="csv"):
        """Load forecast data."""
        self._df_ambient = pd.read_csv(
            os.path.join(GlobalSettings.DATA_FOLDER, "forecast.csv"), index_col=0
        )
        self._df_ambient.index = pd.DatetimeIndex(self._df_ambient.index)
        self._time = [idx.timestamp() for idx in self._df_ambient.index]

    def _compute_load_forecast(self):
        """Compute load forecast."""
        self._df_ambient["Qdot_c"] = 9e2 * (self._df_ambient["T_amb"] - 18.0)
        self._df_ambient.loc[self._df_ambient["Qdot_c"] <= 0.0, ["Qdot_c"]] = 0.0
        self._df_ambient["Qdot_c"] += 1e3

    def _fill_remaining_nan_values(self):
        self._df_ambient.interpolate(method="linear", inplace=True)
        self._df_ambient.bfill(inplace=True)

    def _manual_tuning_of_forecasts(self):
        self._df_ambient[["I_vtsc", "I_fpsc"]] *= 0.9
        self._df_ambient["T_amb"] += 3

    def get_t0(self):
        """Get start time."""
        return self._df_ambient.index[6]

    def _set_time_steps(self):
        # TODO make it general, based on csv file!
        self.time_steps = [
            (
                timedelta(seconds=self.timing.dt_day)
                if idx < self.timing.N_day + 2
                else timedelta(seconds=self.timing.dt_night)
            )
            for idx in range(self.timing.N)
        ]

    def interpolate(self, value: datetime):
        """Interpolate a value."""
        return {
            k: np.interp(value.timestamp(), self._time, self._df_ambient[k])
            for k in self._df_ambient.keys()
        }


if __name__ == "__main__":
    from matplotlib import pyplot as plt

    timing = Timing()
    ambient = Ambient(timing=timing)

    interp_df = []
    interp_df.append(ambient._df_ambient.loc[ambient.get_t0()].to_dict())
    tmp = timedelta(0)
    for i in range(len(ambient.time_steps)):
        tmp += ambient.time_steps[i]
        interp_df.append(ambient.interpolate(ambient.get_t0() + tmp))
    interp_df = pd.DataFrame(interp_df)

    time_grid = [0]
    for i in range(timing.N):
        time_grid.append(time_grid[i] + ambient.time_steps[i].total_seconds())
    interp_df["time_grid"] = time_grid
    print(interp_df)

    plt.figure()
    plt.plot(range(ambient.timing.N + 1), interp_df.iloc[:, 0])
    plt.plot(range(ambient.timing.N + 1), interp_df.iloc[:, 1])
    plt.plot(range(ambient.timing.N + 1), interp_df.iloc[:, 2])
    plt.plot(range(ambient.timing.N + 1), interp_df.iloc[:, 3])
    plt.plot(range(ambient.timing.N + 1), interp_df.iloc[:, 4])
    plt.plot(range(ambient.timing.N + 1), interp_df.iloc[:, 5])
    plt.yscale("log")
    plt.show()
    # plt.figure()
    # ax1 = plt.gca()
    # ax1.plot(interp_df.time_grid, interp_df.I_vtsc)
    # ax1.plot(interp_df.time_grid, interp_df.I_fpsc)

    # # ax1.legend(loc="upper left")

    # ax2 = ax1.twinx()
    # ax2.plot(interp_df.time_grid, interp_df.T_amb)
    # ax2.legend(loc="upper right")

    # plt.show()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""MIOCP specific solver based on the Combinatorial Integral Approximation with Pycombina."""

import casadi as ca
import logging
from typing import Tuple
from camino.data import MinlpData
from camino.problem import MinlpProblem
from camino.settings import Settings
from camino.solvers import MiSolverClass
from camino.solvers.approximation.cia import PycombinaSolver
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.stats import Stats
from camino.utils import colored, toc
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class CiaSolver(MiSolverClass):

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, settings: Settings
    ):
        super(CiaSolver, self).__init__(problem, data, stats, settings)
        logger.info("Setup NLP solver and Pycombina...")
        self.nlp = NlpSolver(problem, self.stats, self.settings)
        self.combina_solver = PycombinaSolver(problem, self.stats, self.settings)
        logger.info("Solver initialized.")

    def solve(self, nlpdata: MinlpData) -> MinlpData:

        toc()
        # Solve relaxed NLP(y^k)
        nlpdata = self.nlp.solve(nlpdata, integers_relaxed=True)
        self.stats["iter_nr"] = 0
        self.stats["best_iter"] = 0
        self.stats["nlp_obj"] = nlpdata.obj_val
        self.stats["lb"] = nlpdata.obj_val
        self.stats["x_sol"] = to_0d(nlpdata.x_sol)
        self.stats["nlp_rel_time"] = toc()
        if not nlpdata.solved:
            logger.error(colored("Relaxed NLP not solved."))
            raise RuntimeError()

        if self.settings.WITH_LOG_DATA:
            self.stats.save()

        # Solve CIA problem
        nlpdata = self.combina_solver.solve(nlpdata)

        # Solve NLP with fixed integers
        nlpdata = self.nlp.solve(nlpdata, integers_relaxed=False)
        self.stats["iter_nr"] = 1
        self.stats["best_iter"] = 1
        self.stats["nlp_obj"] = nlpdata.obj_val
        # self.stats["lb"] = nlpdata.obj_val
        self.stats["x_sol"] = to_0d(nlpdata.x_sol)
        self.stats["total_time_calc"] = toc(reset=True)
        if not nlpdata.solved:
            logger.error(colored("NLP with fixed binary not solved."))
            raise RuntimeError()
        self.update_best_solutions(nlpdata)
        return self.get_best_solutions(nlpdata)

    def reset(self, nlpdata: MinlpData):
        logger.warning(colored("Nothing to reset for CIA solver.", "yellow"))
        return nlpdata

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart."""
        if not nlpdata.relaxed:
            self.update_best_solutions(nlpdata)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""An NLP solver."""

import casadi as ca
import numpy as np
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import GlobalSettings, Settings
from camino.utils.conversion import to_0d


class FeasibilityNlpSolver(SolverClass):
    """Create solver for feasible NLP."""

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create solver for feasible NLP."""
        super(FeasibilityNlpSolver, self).__init__(problem, stats, s)
        options = regularize_options(s.IPOPT_SETTINGS, {}, s)

        g = []
        self.g_idx = []
        beta = GlobalSettings.CASADI_VAR.sym("beta")
        for i in range(problem.g.shape[0]):
            if data.lbg[i] == data.ubg[i]:
                # when lbg == ubg we have an equality constraints, so we need to append it only once
                if abs(data.lbg[i]) == np.inf:
                    raise ValueError(
                        "lbg and ubg cannot be +-inf at the same time, "
                        "at instant {i}, you have {data.lbg[i]=} and {data.ubg[i]=}"
                    )
                g.append(-problem.g[i] + data.lbg[i] - beta)
                self.g_idx.append((i, 1))
            else:
                if data.lbg[i] != -np.inf:
                    g.append(-problem.g[i] + data.lbg[i] - beta)
                    self.g_idx.append((i, -1))
                if data.ubg[i] != np.inf:
                    g.append(problem.g[i] - data.ubg[i] - beta)
                    self.g_idx.append((i, 1))
        g = ca.vertcat(*g)
        self.nr_g = g.shape[0]

        self.lbg = -np.inf * np.ones((self.nr_g))
        self.ubg = np.zeros((self.nr_g))

        if problem.idx_g_dwelltime is not None:
            new_g_constraint = problem.g[problem.idx_g_without_dwelltime]
        else:
            new_g_constraint = problem.g
        self.idx_g_without_dwelltime = problem.idx_g_without_dwelltime

        self.lbg = self.lbg[self.idx_g_without_dwelltime].flatten()
        self.ubg = self.ubg[self.idx_g_without_dwelltime].flatten()

        f = beta
        x = ca.vertcat(*[problem.x, beta])
        p = ca.vertcat(*[problem.p])
        self.idx_x_integer = problem.idx_x_integer
        options.update({"jit": s.WITH_JIT})
        self.solver = ca.nlpsol(
            "nlpsol", "ipopt", {"f": f, "g": new_g_constraint, "x": x, "p": p}, options
        )

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """solve."""
        print("FEASIBILITY")
        success_out = []
        sols_out = []
        # add lower bound for the slacks
        lbx = ca.vcat([nlpdata.lbx, np.zeros(1)])
        # add upper bound for the slacks
        ubx = ca.vcat([nlpdata.ubx, np.inf * np.ones(1)])

        for success_prev, sol in zip(nlpdata.solved_all, nlpdata.solutions_all):
            if success_prev:
                success_out.append(success_prev)
                sols_out.append(sol)
            else:
                lbx[self.idx_x_integer] = to_0d(sol["x"][self.idx_x_integer])
                ubx[self.idx_x_integer] = to_0d(sol["x"][self.idx_x_integer])

                sol_new = self.solver(
                    x0=ca.vcat(
                        [
                            nlpdata.x_sol[: nlpdata.x0.shape[0]],
                            np.zeros(1),
                        ]  # slacks initialization
                    ),
                    lbx=lbx,
                    ubx=ubx,
                    lbg=self.lbg,
                    ubg=self.ubg,
                    p=nlpdata.p,
                )

                # Reconstruct lambda_g:
                lambda_g = to_0d(sol_new["lam_g"])
                lambda_g_req = np.zeros(nlpdata.lbg.shape[0])
                for lg, (idx, sgn) in zip(lambda_g, self.g_idx):
                    lambda_g_req[idx] = sgn * lg
                sol_new["lam_g"] = ca.DM(lambda_g_req)

                success, _ = self.collect_stats("F-NLP", sol=sol_new)
                if not success:
                    print("FNLP not solved")
                    raise Exception("FNLP not solved!")

                # Maintain that it is caused due to infeasibility!!!
                success_out.append(False)
                sols_out.append(sol_new)

        nlpdata.prev_solutions = sols_out
        nlpdata.solved_all = success_out
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Utilities for pump algorithms."""

import numpy as np
from random import randint, random
from camino.data import MinlpData
from camino.utils import logging
from camino.utils.conversion import to_0d
from copy import deepcopy

logger = logging.getLogger(__name__)


def integer_error(x_int, norm=1):
    """Compute integer error."""
    if norm == 1:
        ret = np.sum(np.abs(np.round(x_int) - x_int))
    else:
        ret = np.linalg.norm(np.round(x_int) - x_int)

    logger.info(f"Integer error {ret:.3f} / {x_int.shape[0]:.3f}")
    return ret


def create_rounded_data(data: MinlpData, idx_x_integer):
    for i in range(data.nr_sols):
        # Round the continuous solution
        x_var = to_0d(data.prev_solutions[i]["x"])
        x_var[idx_x_integer] = np.round(x_var[idx_x_integer])
        datarounded = deepcopy(data)
        datarounded.prev_solutions[i]["x"] = x_var
    return datarounded


def perturbe_x(x_current, idx_x_integer):
    """
    Perturbe x as described in:

        Bertacco, L., Fischetti, M., & Lodi, A. (2007). A feasibility pump heuristic for general mixed-integer
        problems. Discrete Optimization, 4(1), 63-76.

    For TT integer variables with largest integer difference, round in the other direction if the fractional difference
    is large than 0.02.
    """
    N = len(idx_x_integer)
    T = N // 10  # TunedParameter
    TT = min(randint(T // 2, (T * 3) // 2), N)
    x_bin = x_current[idx_x_integer]
    x_rounded = np.round(x_bin)
    x_diff = np.abs(x_bin - x_rounded)
    idx_largest = np.argpartition(x_diff, -TT)[-TT:]

    for i in idx_largest:
        if x_rounded[i] < x_bin[i]:
            x_rounded[i] += 1
        else:
            x_rounded[i] -= 1

    x_current[idx_x_integer] = x_rounded
    return x_current


def randrange(xmin, xmax):
    """Rand range."""
    return random() * (xmax - xmin) + xmin


def random_perturbe_x(x_current, idx_x_integer):
    """
    Random perturbation of x as described in:

        Bertacco, L., Fischetti, M., & Lodi, A. (2007). A feasibility pump heuristic for general mixed-integer problems.
        Discrete Optimization, 4(1), 63-76.
    """
    N = len(idx_x_integer)
    random_range = [-0.3, 0.7]
    x_bin = x_current[idx_x_integer]
    x_rounded = np.round(x_bin)
    x_diff = np.abs(x_bin - x_rounded)

    for i in range(N):
        if x_diff[i] + max(randrange(*random_range), 0) > 0.5:
            if x_rounded[i] < x_bin[i]:
                x_rounded[i] += 1
            else:
                x_rounded[i] -= 1

    x_current[idx_x_integer] = x_rounded
    return x_current



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import casadi as ca
from camino.settings import GlobalSettings
from camino.problems import MetaDataOcp
from camino.problems.dsc import Description
from camino.utils.integrators import integrate_rk4


def time_opt_car():
    """Time optimal car with a goal."""
    v_turbo = 10
    v_max = 100
    N = 20
    p_des = 500
    v_des = 0.0

    p = GlobalSettings.CASADI_VAR.sym("p")
    v = GlobalSettings.CASADI_VAR.sym("v")
    t = GlobalSettings.CASADI_VAR.sym("t")
    x = ca.vcat([p, v, t])
    x_lb = [0, 0, 0]
    x_ub = [ca.inf, v_max, ca.inf]
    x_0 = [0, 0, 0]
    x_discrete = [0, 0, 0]

    # Controls
    a = GlobalSettings.CASADI_VAR.sym("a")
    turbo = GlobalSettings.CASADI_VAR.sym("turbo")  # Boolean
    reverse = GlobalSettings.CASADI_VAR.sym("reverse")  # gear
    u = ca.vcat([a, turbo, reverse])
    u_lb = [0, 0, -1]
    u_ub = [10, 1, 1]
    u_0 = [0, 0, 0]
    u_discrete = [0, 1, 1]
    u_path_weights = [0.5, 0, 0]

    # Dynamics
    x_dot = ca.vertcat(v, a * (1 + turbo) * reverse, 1)

    # Path constraints 0 < constraint
    p_g = ca.Function(
        "g",
        [x, u],
        [
            v / v_turbo - turbo,  # Turbo only on at v > v_turbo,
        ],
    )
    p_g_lb = 0
    p_g_ub = ca.inf

    terminal_g = ca.Function("g_term", [x], [ca.vertcat(p - p_des, v - v_des)])
    terminal_g_lb = 0
    terminal_g_ub = 0

    # Terminal cost
    stage_cost = ca.Function("stage_cost", [x, u], [0])
    terminal_cost = ca.Function("T", [x], [t])

    # Discretize
    dt = GlobalSettings.CASADI_VAR.sym("dt")
    dt_lb = 0.0
    dt_ub = 1.0
    F = integrate_rk4(x, u, x_dot, dt)

    # Create OCP
    dsc = Description()
    xp = dsc.add_parameters("x0", x.numel(), x_0)
    up = dsc.add_parameters("u0", u.numel(), u_0)
    dtk = dsc.sym("dt", 1, dt_lb, dt_ub)
    for i in range(N):
        xk = dsc.sym("x", x.numel(), x_lb, x_ub, x_0, x_discrete)
        uk = dsc.sym("u", u.numel(), u_lb, u_ub, discrete=u_discrete)
        dsc.eq(F(xp, uk, dtk), xk)
        dsc.add_g(p_g_lb, p_g(xk, uk), p_g_ub)
        dsc.f += dtk * stage_cost(xk, uk)
        dsc.f += sum(
            [v * (uk[id] - up[id]) ** 2 for id, v in enumerate(u_path_weights)]
        )
        xp, up = xk, uk
    # Final cost
    dsc.add_g(terminal_g_lb, terminal_g(xp), terminal_g_ub)
    dsc.f += terminal_cost(xp)

    problem = dsc.get_problem()
    problem.meta = MetaDataOcp(
        dt=None,
        initial_state=x_0[:-1],
        n_state=x.numel() - 1,
        n_continuous_control=u.numel(),
        n_discrete_control=0,
        idx_state=np.vstack(dsc.get_indices("x"))[:, :-1],
        idx_control=np.vstack(dsc.get_indices("u")),
        idx_bin_control=[],
        idx_t=np.vstack(dsc.get_indices("x"))[:, -1],
    )
    problem.hessian_not_psd = True
    data = dsc.get_data()
    return problem, data



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Overview of all problems."""

import logging
from camino.settings import Settings, GlobalSettings
from camino.problems import MinlpProblem, MinlpData, MetaDataOcp
import casadi as ca
import numpy as np
from camino.problems.dsc import Description
from camino.solvers import Stats
from camino.utils.integrators import integrate_rk4  # integrate_ee
from camino.problems.double_tank import create_double_tank_problem2
from camino.problems.solarsys import create_stcs_problem
from camino.problems.gearbox import (
    create_simple_gearbox,
    create_gearbox,
    create_gearbox_int,
)
from camino.problems.load_problem_from_nl_file import create_from_nl_file
from camino.problems.manual_formulation_from_minlplib import alan
from camino.utils.conversion import to_bool
from camino.problems.time_opt import time_opt_car
from camino.problems.sto_based import particle_trajectory

logger = logging.getLogger(__name__)


def create_ocp_unstable_system(p_val=[0.9, 0.7]):
    """
    OCP of a unstable system subject to min uptime constraints.

    Example taken from preprint of A. Buerger. Inspired by a textbook example of the MPC book by Rawlings, Mayne, Diehl
    """
    dt = 0.05
    N = 30
    min_uptime = 2  # in time steps

    dsc = Description()
    x = GlobalSettings.CASADI_VAR.sym("x")  # state
    u = GlobalSettings.CASADI_VAR.sym("u")  # control
    Xref = dsc.add_parameters("Xk_ref", 1, p_val[1])

    xdot = x**3 - u
    r = ca.Function("residual_obj", [x, Xref], [x - Xref])
    F = integrate_rk4(x, u, xdot, dt, m_steps=1)

    Xk = dsc.add_parameters("Xk0", 1, p_val[0])
    BigM = 1e2
    Uprev = None
    for k in range(N):
        Uk = dsc.sym("Uk", 1, lb=0, ub=1, w0=1, discrete=True)

        if Uprev is not None and min_uptime > 0:
            uptime_step = 1
            it = 0
            while uptime_step < min_uptime:
                if uptime_step < min_uptime:
                    idx_1 = k - 1
                    idx_2 = k - (it + 2)

                    if idx_1 >= 0:
                        b_idx_1 = int(np.array(dsc.get_indices("Uk")[idx_1]))
                        b_idx_1 = dsc.w[b_idx_1]
                    else:
                        b_idx_1 = 0

                    if idx_2 >= 0:
                        b_idx_2 = int(np.array(dsc.get_indices("Uk")[idx_2]))
                        b_idx_2 = dsc.w[b_idx_2]
                    else:
                        b_idx_2 = 0
                    dsc.leq(-Uk + b_idx_1 - b_idx_2, 0, is_dwell_time=1)
                    it += 1
                    uptime_step += 1
                    logger.debug(dsc.g[-1])

        # Integrate till the end of the interval
        Xk_end = F(Xk, Uk)
        dsc.f += (Xk - Xref) ** 2
        dsc.r += [r(Xk, Xref)]

        # New NLP variable for state at end of interval
        Xk = dsc.sym("Xk", 1, lb=-BigM, ub=BigM, w0=p_val[0])
        dsc.eq(Xk_end, Xk)

        Uprev = Uk
    dsc.f += (Xk - Xref) ** 2
    problem = dsc.get_problem()
    meta = MetaDataOcp(
        N_horizon=N,
        dt=dt,
        n_state=1,
        n_discrete_control=1,
        n_continuous_control=0,
        initial_state=p_val[0],
        idx_control=np.hstack(dsc.get_indices("Uk")),
        idx_state=np.hstack(dsc.get_indices("Xk")),
        idx_bin_control=problem.idx_x_integer,
        scaling_coeff_control=[1],
        min_uptime=min_uptime,
        f_dynamics=F,
    )
    problem.meta = meta
    data = dsc.get_data()
    s = Settings()
    s.USE_RELAXED_SOL_AS_LINEARIZATION = True
    s.USE_TIGHT_MIPGAP_FIRST_ITERATION = True
    s.CONSTRAINT_INT_TOL = 1e-5
    s.MINLP_TOLERANCE_ABS = 1e-4
    s.MINLP_TOLERANCE = 1e-4
    s.IPOPT_SETTINGS.update(
        {
            "ipopt.linear_solver": "ma27",
            "ipopt.max_iter": 1000,
        }
    )
    s.MIP_SETTINGS_ALL["gurobi"].update(
        {
            "gurobi.FeasibilityTol": s.CONSTRAINT_INT_TOL,
            "gurobi.IntFeasTol": s.CONSTRAINT_INT_TOL,
            #  "gurobi.Threads": 1,
            "gurobi.PoolSearchMode": 0,
            "gurobi.PoolSolutions": 3,
            # "gurobi.NodeMethod": 2,
            # "gurobi.Method": 2,
        }
    )
    s.BONMIN_SETTINGS.update(
        {
            "bonmin.allowable_fraction_gap": s.MINLP_TOLERANCE,
            "bonmin.allowable_gap": s.MINLP_TOLERANCE_ABS,
            "bonmin.linear_solver": "ma27",
        }
    )
    return problem, data, s


def create_check_sign_lagrange_problem():
    """Create a problem to check the sign of the multipliers."""
    x = GlobalSettings.CASADI_VAR.sym("x")
    p = GlobalSettings.CASADI_VAR.sym("p")

    problem = MinlpProblem(x=x, f=(x - 2) ** 2, g=ca.vcat([x]), p=p, idx_x_integer=[0])
    data = MinlpData(
        x0=np.array([0]),
        _ubx=np.array([np.inf]),
        _lbx=np.array([-np.inf]),
        _ubg=np.array([-1]),
        _lbg=np.array([-7]),
        p=np.array([0]),
    )

    return problem, data


def create_dummy_problem(p_val=[1000, 3]):
    """
    Create a dummy problem.

    This problem corresponds to the tutorial example in the GN-Voronoi paper.
    (apart from the upper bound)
    """
    opti = Description()
    x0 = np.array([0, 4, 100])
    lbx = np.array([0, 0, 0])
    ubx = np.array([4, 4, np.inf])
    x = opti.sym("x", shape=3, lb=lbx, ub=ubx, w0=x0, discrete=[1, 1, 0])
    # idx_x_integer = [0, 1]
    p = opti.add_parameters("p", shape=2, values=p_val)
    opti.f = (x[0] - 4.1) ** 2 + (x[1] - 4.0) ** 2 + x[2] * p[0]
    ubg = np.array([ca.inf, ca.inf])
    lbg = np.array([0, 0])
    opti.add_g(
        lbg,
        ca.vertcat(x[2], -(x[0] ** 2 + x[1] ** 2 - x[2] - p[1] ** 2)),
        ubg,
        is_linear=1,
    )
    problem = opti.get_problem()
    data = opti.get_data()
    return problem, data


def create_dummy_problem_2():
    """Create a dummy problem."""
    x = GlobalSettings.CASADI_VAR.sym("x", 2)
    x0 = np.array([2, 4])
    idx_x_integer = [0]
    p = GlobalSettings.CASADI_VAR.sym("p", 1)
    f = x[0] ** 2 + x[1]
    g = ca.vertcat(x[1], -x[0] ** 2 - x[1] / 4 + p[0] ** 2)
    ubg = np.array([np.inf, np.inf])
    lbg = np.array([-3.0, 0.0])
    lbx = -np.inf * np.ones((2,))
    ubx = np.array([ca.inf, ca.inf])

    problem = MinlpProblem(x=x, f=f, g=g, p=p, idx_x_integer=idx_x_integer)
    data = MinlpData(x0=x0, _ubx=ubx, _lbx=lbx, _ubg=ubg, _lbg=lbg, p=[100])
    return problem, data


def create_double_pipe_problem(p_val=[1, 5, 1, 10]):
    """Create double pipe problem."""
    y = GlobalSettings.CASADI_VAR.sym("y", 1)  # integers
    z = GlobalSettings.CASADI_VAR.sym("z", 2)  # continuous
    x0 = np.array([1, 0, 0])
    x = ca.vertcat(*[y, z])
    idx_x_integer = [0]

    alpha = GlobalSettings.CASADI_VAR.sym("alpha", 2)
    r = GlobalSettings.CASADI_VAR.sym("r", 1)
    gamma = GlobalSettings.CASADI_VAR.sym("gamma", 1)
    p = ca.vertcat(*[alpha, r, gamma])

    f = alpha[0] * z[0] + alpha[1] * y[0]
    g = ca.vertcat(*[z[0] + y[0] * z[1] - r, gamma - z[1]])
    lbg = np.array([0, 0])
    ubg = np.array([ca.inf, ca.inf])
    lbx = np.array([0, 0, -ca.inf])
    ubx = np.array([1, ca.inf, ca.inf])

    problem = MinlpProblem(x=x, f=f, g=g, p=p, idx_x_integer=idx_x_integer)
    data = MinlpData(x0=x0, _ubx=ubx, _lbx=lbx, _ubg=ubg, _lbg=lbg, p=p_val)
    return problem, data


def create_double_tank_problem(p_val=[2, 2.5]):
    """
    Implement the double tank problem.

    Taken from Abbasi et al. ECC 23, reimplemented to achieve nice sparsity pattern.

    """
    eps = 1e-3
    N = 300  # NOTE In paper the is set to 300
    dt = 1 / 30
    T = N * dt
    alpha = 100
    beta = np.array([1.0, 1.2])
    gamma = 10
    scaling_coeff = [gamma, 1]
    demand = np.array([2 + 0.5 * np.sin(x) for x in np.linspace(0, T, N + 1)])

    nx = 2
    nu = 2
    x_0 = GlobalSettings.CASADI_VAR.sym("x0", nx)
    x = GlobalSettings.CASADI_VAR.sym("x", nx)  # state
    u = GlobalSettings.CASADI_VAR.sym("u", nu)  # control
    x1dot = scaling_coeff[0] * u[0] + u[1] - ca.sqrt(x[0] + eps)
    x2dot = ca.sqrt(x[0] + eps) - ca.sqrt(x[1] + eps)
    xdot = ca.vertcat(*[x1dot, x2dot])

    # F = integrate_ee(x, u, xdot, dt, m_steps=1)
    F = integrate_rk4(x, u, xdot, dt, m_steps=1)

    w = []
    w0 = []
    lbw = []
    ubw = []
    J = 0
    g = []
    lbg = []
    ubg = []
    p = []
    idx_x_integer = []
    idx_state = []
    idx_control = []
    idx_var = 0

    Xk = x_0
    p += [Xk]
    for k in range(N):
        Uk = GlobalSettings.CASADI_VAR.sym(f"u_{k}", nu)
        w += [Uk]
        idx_var += nu
        lbw += [0, 0]
        ubw += [1, gamma]
        w0 += [0.5, 0.5]
        idx_x_integer.append(np.arange(idx_var - nu, idx_var - 1))
        idx_control.append(np.arange(idx_var - nu + 1, idx_var))

        # Integrate till the end of the interval
        Xk_end = F(Xk, Uk)
        J += dt * alpha * (Xk[1] - demand[k]) ** 2
        J += dt * (beta[0] * scaling_coeff[0] * Uk[0] + beta[1] * Uk[1])

        # New NLP variable for state at end of interval
        Xk = GlobalSettings.CASADI_VAR.sym(f"x_{k+1}", nx)
        idx_var += nx
        idx_state.append(np.arange(idx_var - nx, idx_var))
        w += [Xk]
        lbw += [0, 0]
        ubw += [1e3, 1e3]
        w0 += [0.5, 0.5]

        # Add equality constraint
        g += [Xk_end - Xk]
        lbg += [0, 0]
        ubg += [0, 0]

    meta = MetaDataOcp(
        dt=dt,
        n_state=nx,
        n_continuous_control=1,
        n_discrete_control=1,
        initial_state=p_val,
        idx_control=np.hstack(idx_control),
        idx_bin_control=np.hstack(idx_x_integer),
        idx_state=np.hstack(idx_state),
        scaling_coeff_control=scaling_coeff,
    )
    problem = MinlpProblem(
        x=ca.vcat(w),
        f=J,
        g=ca.vcat(g),
        p=x_0,
        idx_x_integer=np.hstack(idx_x_integer),
        meta=meta,
    )
    data = MinlpData(
        x0=ca.vcat(w0),
        _ubx=ca.vcat(ubw),
        _lbx=ca.vcat(lbw),
        _ubg=np.array(ubg),
        _lbg=np.array(lbg),
        p=p_val,
    )
    return problem, data


def counter_example_nonconvexity():
    """Nonconvexity example."""
    x = GlobalSettings.CASADI_VAR.sym("x")
    y = GlobalSettings.CASADI_VAR.sym("y")

    f = ca.atan(x - 0.3) ** 2 + x / 10 + x**2 / 50 + y**2
    problem = MinlpProblem(
        x=ca.vcat([x, y]), f=f, g=ca.MX([]), p=ca.MX([]), idx_x_integer=[0]
    )
    data = MinlpData(
        x0=np.array([-4, 2]),
        _lbx=np.array([-5, -5]),
        _ubx=np.array([5, 5]),
        _ubg=[],
        _lbg=[],
        p=[],
    )
    return problem, data


def reduce_list(data):
    """Reduce list."""
    out = []
    for el in data:
        if isinstance(el, list):
            out.extend(reduce_list(el))
        else:
            out.append(el)
    return out


def create_from_nosnoc(file, compiled=False):
    """Create a problem from nosnoc."""
    from camino.utils.data import load_pickle
    from camino.utils.cache import CachedFunction, cache_data, return_func
    from camino.solvers import get_lin_bounds
    from os import path

    name = path.basename(file)
    data = load_pickle(file)
    x = GlobalSettings.CASADI_VAR.sym("x", data["w_shape"][0])
    p = GlobalSettings.CASADI_VAR.sym("p", data["p_shape"][0])
    if to_bool(compiled):
        f = CachedFunction(f"{name}_f", return_func(data["f"]))(x, p)
        g = CachedFunction(f"{name}_g", return_func(data["g"]))(x, p)
    else:
        f = data["f"](x, p)
        g = data["g"](x, p)

    print("Loaded Functions")
    idx_x_integer = reduce_list(data["ind_bool"])
    problem = MinlpProblem(
        x=x, p=p, f=f, g=g, idx_x_integer=idx_x_integer, hessian_not_psd=True
    )
    if to_bool(compiled):
        problem.idx_g_lin, problem.idx_g_lin_bin = cache_data(
            f"{name}_id", get_lin_bounds, problem
        )
    else:
        # Probably this is just overkill, set them to 0
        problem.idx_g_lin, problem.idx_g_lin_bin = [], []

    data["w0"][idx_x_integer] = np.round(data["w0"][idx_x_integer])
    data = MinlpData(
        p=data["p0"],
        x0=data["w0"],
        _lbx=data["lbw"],
        _ubx=data["ubw"],
        _lbg=data["lbg"],
        _ubg=data["ubg"],
    )
    return problem, data


def create_from_sto(file, with_uptime=True):
    """Create a problem from STO."""
    from camino.utils.data import load_pickle
    from camino.utils.conversion import to_bool

    with_uptime = to_bool(with_uptime)
    data = load_pickle(file)
    dt = data["dt"]
    N = data["N"]

    nx = data["lbx"].shape[0] - 1
    p = GlobalSettings.CASADI_VAR.sym("p", 1)
    p0 = [dt * N]
    x = GlobalSettings.CASADI_VAR.sym("x", nx)
    x0 = data["init"][1:]
    ubx, lbx = data["ubx"][1:], data["lbx"][1:]
    f = data["f"](ca.vertcat(p, x))
    g = data["g"](ca.vertcat(p, x))
    idx_x_integer = np.where(data["int"][1:].full() == 1)[0].tolist()
    ub_idx = data["ub_idx"] - 1
    u_idx = data["u_idx"] - 1
    x_idx = data["x_idx"] - 1

    # Add extra switching constraints:
    switches = []
    constraints_eq = []  # 0 < item
    constraints_leq = []  # 0 < item
    min_uptime = [int(x / dt) for x in data["min_uptime"]]
    min_downtime = [int(x / dt) for x in data["min_downtime"]]
    ub_idx_c = ub_idx.reshape((-1, data["nub"]))
    if with_uptime:
        for i, (min_up, min_down) in enumerate(zip(min_uptime[:-1], min_downtime[:-1])):
            if min_up > 1 or min_down > 1:
                switch = GlobalSettings.CASADI_VAR.sym(f"switch_x{i}", N)
                switches.append(switch)
                for idxt in range(N - 1):
                    constraints_eq.append(
                        (x[ub_idx_c[idxt, i]] + switch[idxt]) - x[ub_idx_c[idxt + 1, i]]
                    )

                if min_up > 1:
                    for idxt in range(N - 1):
                        for idxt2 in range(idxt + 1, min(N, idxt + min_up)):
                            # switch < x -> 0 < x - switch
                            constraints_leq.append(x[ub_idx_c[idxt2, i]] - switch[idxt])

                if min_down > 1:
                    for idxt in range(N - 1):
                        for idxt2 in range(idxt + 1, min(N, idxt + min_down)):
                            # 1 + switch < 1 - x -> 0 < 1 - x - switch - 1
                            constraints_leq.append(x[ub_idx_c[idxt2, i]] + switch[idxt])
    # Add to problem!
    nr_switches = len(switches)
    switches = ca.vcat(switches)
    constraints_eq = ca.vcat(constraints_eq)
    constraints_leq = ca.vcat(constraints_leq)
    nub = data["nub"] + nr_switches
    if nr_switches > 0:
        idx_switches = [
            [nx + N * swi + i for i in range(N)] for swi in range(nr_switches)
        ]
        ub_idx_c = np.hstack((ub_idx_c, np.array(idx_switches).T))
        min_uptime = min_uptime[:-1] + [0] * nr_switches + [min_uptime[-1]]
        min_downtime = min_downtime[:-1] + [0] * nr_switches + [min_downtime[-1]]

    ub_idx_c = ub_idx_c.reshape((-1,))
    x_total = ca.vertcat(x, switches)
    g = ca.vertcat(g, constraints_eq, constraints_leq)
    lbx = ca.vertcat(lbx, -np.ones(switches.shape))
    ubx = ca.vertcat(ubx, np.ones(switches.shape))
    x0 = ca.vertcat(x0, np.zeros(switches.shape))
    x0[idx_x_integer] = np.round(x0[idx_x_integer])
    lbg = ca.vertcat(
        data["lbg"], np.zeros(constraints_eq.shape), np.zeros(constraints_leq.shape)
    )
    ubg = ca.vertcat(
        data["ubg"],
        np.zeros(constraints_eq.shape),
        ca.inf * np.ones(constraints_leq.shape),
    )

    problem = MinlpProblem(
        x=x_total, p=p, f=f, g=g, idx_x_integer=idx_x_integer, hessian_not_psd=True
    )
    problem.idx_g_lin, problem.idx_g_lin_bin = [], []

    meta = MetaDataOcp(
        dt=data["dt"],
        n_state=data["nx"],
        n_continuous_control=data["nu"] - data["nub"],
        n_discrete_control=nub,
        initial_state=data["x0"],
        idx_control=u_idx,
        idx_state=x_idx,
        idx_bin_control=ub_idx_c,
        min_uptime=min_uptime,
        min_downtime=min_downtime,
        dump_solution=ca.Function("dump", [x_total], [ca.vertcat(p0, x)]),
    )

    problem.meta = meta

    data = MinlpData(p=p0, x0=x0, _lbx=lbx, _ubx=ubx, _lbg=lbg, _ubg=ubg)
    return problem, data


def create_from_nlpsol_description(file):
    """Create a problem from casadi nlpsol description."""
    from camino.utils.data import load_pickle

    data = load_pickle(file)

    x = GlobalSettings.CASADI_VAR.sym("x", data["nx"])
    p = GlobalSettings.CASADI_VAR.sym("p", data["np"])
    f = data["f_fun"](x)
    g = data["g_fun"](x)
    idx_x_integer = data["idx_x_integer"]

    problem = MinlpProblem(
        x=x, p=p, f=f, g=g, idx_x_integer=idx_x_integer, hessian_not_psd=True
    )

    data["x0"][idx_x_integer] = np.round(data["x0"][idx_x_integer])
    data = MinlpData(
        p=data["p0"],
        x0=data["x0"],
        _lbx=data["lbx"],
        _ubx=data["ubx"],
        _lbg=data["lbg"],
        _ubg=data["ubg"],
    )
    return problem, data


PROBLEMS = {
    "sign_check": create_check_sign_lagrange_problem,
    "dummy": create_dummy_problem,
    "dummy2": create_dummy_problem_2,
    "doublepipe": create_double_pipe_problem,
    "doubletank": create_double_tank_problem,
    "doubletank2": create_double_tank_problem2,
    "stcs": create_stcs_problem,
    "stcs-simplified": lambda *args, **kwargs: create_stcs_problem(
        *args, **kwargs, simplified=True
    ),
    "gearbox": create_simple_gearbox,
    "gearbox_int": create_gearbox_int,
    "gearbox_complx": create_gearbox,
    "nonconvex": counter_example_nonconvexity,
    "unstable_ocp": create_ocp_unstable_system,
    "nosnoc": create_from_nosnoc,
    "from_sto": create_from_sto,
    "from_nlpsol_dsc": create_from_nlpsol_description,
    "to_car": time_opt_car,
    "particle": particle_trajectory,
    "alan": alan,
    "nl_file": create_from_nl_file,
}


if __name__ == "__main__":
    from camino.solvers.subsolvers.nlp import NlpSolver
    from camino.utils import plot_trajectory
    from camino.utils.conversion import to_0d
    import matplotlib.pyplot as plt
    from camino.solvers.decomposition.voronoi_master import VoronoiTrustRegionMIQP

    stats = Stats(mode="custom", problem_name="unstable_ocp")
    prob, data, settings = create_ocp_unstable_system()

    nlp = NlpSolver(prob, stats, settings)
    data = nlp.solve(data, integers_relaxed=True)
    print(f"Relaxed  {data.obj_val=}")

    # Solve MIQP around MINLP solution
    miqp = VoronoiTrustRegionMIQP(prob, data, stats, settings)
    data = miqp.solve(data, prev_feasible=True, is_qp=True)

    data = nlp.solve(data, integers_relaxed=False)
    x_star = data.x_sol
    print(f"{data.obj_val=}")

    if isinstance(prob.meta, MetaDataOcp):
        meta = prob.meta
        state = to_0d(x_star)[meta.idx_state].reshape(-1, meta.n_state)
        state = np.vstack([meta.initial_state, state])
        control = to_0d(x_star)[meta.idx_control].reshape(-1, meta.n_continuous_control)
        fig, axs = plot_trajectory(
            to_0d(x_star), state, control, meta, title="problem name"
        )
        plt.show()

    # grad_f = ca.Function('grad_f', [prob.x, prob.p], [ca.gradient(prob.f, prob.x)])
    # grad_g = ca.Function('grad_g', [prob.x, prob.p], [ca.jacobian(prob.g, prob.x).T])
    # lambda_k = grad_f(data.x_sol, data.p) + grad_g(data.x_sol, data.p) @ data.lam_g_sol
    # assert np.allclose(-data.lam_x_sol, lambda_k)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""A set of solvers that compute relaxed solutions (non-integer feasible)."""

import logging
from camino.data import MinlpData
from camino.solvers import MiSolverClass
from camino.solvers.subsolvers.nlp import NlpSolver as SubSolver
from camino.utils import colored
from camino.utils.validate import check_integer_feasible

logger = logging.getLogger(__name__)


class NlpSolver(MiSolverClass):

    def __init__(self, problem, data, stats, settings, integers_relaxed=False):
        super(NlpSolver, self).__init__(problem, data, stats, settings)
        self.nlp = SubSolver(problem, stats, settings)
        self.integers_relaxed = integers_relaxed

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        nlpdata = self.nlp.solve(nlpdata, integers_relaxed=self.integers_relaxed)
        nlpdata.relaxed = check_integer_feasible(
            self.nlp.idx_x_integer, nlpdata.x_sol, self.settings, throws=False
        )
        if self.integers_relaxed:
            self.stats.relaxed_solution = nlpdata
        return nlpdata

    def reset(self, nlpdata: MinlpData):
        logger.warning(colored("Nothing to reset for RelaxedNlpSolver"))

    def warmstart(self, nlpdata: MinlpData):
        pass



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Load a MINLP from a nl file."""

from camino.settings import Settings, GlobalSettings
from camino.problems import MinlpProblem, MinlpData, MetaDataOcp
import casadi as ca
import numpy as np


def create_from_nl_file(file, compiled=True):
    """Load from NL file."""
    from camino.utils.cache import CachedFunction, return_func
    import hashlib

    # Create an NLP instance
    nl = ca.NlpBuilder()

    # Parse an NL-file
    nl.import_nl(file, {"verbose": False})
    print(f"Loading MINLP with: {nl.repr()}")

    if not isinstance(nl.x[0], GlobalSettings.CASADI_VAR):
        raise Exception(f"Set GlobalSettings.CASADI_VAR to {type(nl.x[0])} in defines!")

    idx = np.where(np.array(nl.discrete))

    if compiled:
        key = str(hashlib.md5(file.encode()).hexdigest())[:64]
        x = ca.vcat(nl.x)
        problem = MinlpProblem(
            x=x,
            f=CachedFunction(f"f_{key}", return_func(ca.Function("f", [x], [nl.f])))(x),
            g=CachedFunction(
                f"g_{key}", return_func(ca.Function("g", [x], [ca.vcat(nl.g)]))
            )(x),
            idx_x_integer=idx[0].tolist(),
            p=[],
        )
    else:
        problem = MinlpProblem(
            x=ca.vcat(nl.x),
            f=nl.f,
            g=ca.vcat(nl.g),
            idx_x_integer=idx[0].tolist(),
            p=[],
        )
    if nl.f.is_constant():
        raise Exception("No objective!")

    problem.hessian_not_psd = True
    data = MinlpData(
        x0=np.array(nl.x_init),
        _lbx=np.array(nl.x_lb),
        _ubx=np.array(nl.x_ub),
        _lbg=np.array(nl.g_lb),
        _ubg=np.array(nl.g_ub),
        p=[],
    )

    from camino.solvers import inspect_problem, set_constraint_types

    set_constraint_types(problem, *inspect_problem(problem, data))
    s = Settings()

    s.OBJECTIVE_TOL = 1e-8
    s.CONSTRAINT_TOL = 1e-8
    s.CONSTRAINT_INT_TOL = 1e-3
    s.MINLP_TOLERANCE = 1e-2
    s.MINLP_TOLERANCE_ABS = 1e-2
    s.BRMIQP_GAP = 0.1
    s.LBMILP_GAP = 0.01
    s.TIME_LIMIT = 300
    s.TIME_LIMIT_SOLVER_ONLY = False
    s.USE_RELAXED_SOL_AS_LINEARIZATION = True
    s.USE_TIGHT_MIPGAP_FIRST_ITERATION = True
    s.IPOPT_SETTINGS = {
        "ipopt.linear_solver": "ma27",
        "ipopt.max_cpu_time": s.TIME_LIMIT / 4,
        "ipopt.max_iter": 1000,
        "ipopt.constr_viol_tol": s.CONSTRAINT_TOL,
        # "ipopt.mu_strategy": "adaptive",
        # "ipopt.mu_oracle": "probing",
        # "ipopt.bound_relax_factor": 0,
        # "ipopt.honor_original_bounds": "yes",
        "ipopt.print_level": 0,
        # # Options used within Bonmin
        # "ipopt.gamma_phi": 1e-8,
        # "ipopt.gamma_theta": 1e-4,
        # "ipopt.required_infeasibility_reduction": 0.1,
        # "ipopt.expect_infeasible_problem": "yes",
        # "ipopt.warm_start_init_point": "yes",
    }
    s.MIP_SETTINGS_ALL["gurobi"] = {
        "gurobi.MIPGap": 0.1,
        "gurobi.FeasibilityTol": s.CONSTRAINT_INT_TOL,
        "gurobi.IntFeasTol": s.CONSTRAINT_INT_TOL,
        "gurobi.Heuristics": 0.05,
        "gurobi.PoolSearchMode": 0,
        "gurobi.PoolSolutions": 5,
        "gurobi.Threads": 1,
        "gurobi.TimeLimit": s.TIME_LIMIT / 2,
        "gurobi.output_flag": 0,
    }
    s.BONMIN_SETTINGS = {
        "bonmin.time_limit": s.TIME_LIMIT,
        "bonmin.tree_search_strategy": "dive",
        "bonmin.node_comparison": "best-bound",
        "bonmin.allowable_fraction_gap": Settings.MINLP_TOLERANCE,
        "bonmin.allowable_gap": Settings.MINLP_TOLERANCE_ABS,
        "bonmin.constr_viol_tol": s.CONSTRAINT_TOL,
        "bonmin.linear_solver": "ma27",
        "bonmin.bound_relax_factor": 1e-14,
        "bonmin.honor_original_bounds": "yes",
    }
    s.WITH_DEBUG = False
    s.WITH_LOG_DATA = False

    return problem, data, s



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""
This file contains the data types related to the problem description.

The MinlpProblem contain the equations, indices of integer variables, special
data to speed up the calculations and meta data.
"""

from dataclasses import dataclass, field
from typing import List, Dict, Optional, Callable, Any
import casadi as ca


@dataclass
class MetaData:
    """Meta data class."""


@dataclass
class MetaDataOcp(MetaData):
    """Meta data in case the problem is an OCP."""

    N_horizon: Optional[int] = None
    n_state: Optional[int] = None
    n_continuous_control: Optional[int] = None
    n_discrete_control: Optional[int] = None
    idx_state: Optional[List[float]] = None
    idx_bin_state: Optional[List[float]] = None
    idx_control: Optional[List[float]] = None
    idx_bin_control: Optional[List[float]] = None
    idx_other: Optional[Dict[str, List[float]]] = None
    idx_param: Optional[dict] = None
    f_dynamics: Optional[Callable] = None
    # TODO: initial_state needs to become an index list of p
    initial_state: Optional[List[float]] = None
    dt: Optional[float] = None
    scaling_coeff_control: Optional[List[float]] = None
    min_uptime: Optional[int] = None
    min_downtime: Optional[int] = None
    dump_solution: ca.Function = None
    idx_t: Optional[int] = None


@dataclass
class MetaDataMpc(MetaData):
    """Meta data in case the problem is an OCP."""

    plot: Callable[[MetaData, ca.DM], bool] = None
    shift: Callable[[MetaData, ca.DM], bool] = None


@dataclass
class MinlpProblem:
    """
    Minlp problem description.

    The variables and equations should be defined
    using the GlobaLSettings.CASADI_VAR.

    You need to at least fill in f, g, x, p and idx_x_integer. Note that if
    your hessian is not positive semi-definite, you might want to set the flag
    to avoid computational problems.
    """

    # Casadi objective expression (1x1)
    f: Any
    # Casadi constraints expression (1xM)
    g: Any
    # Casadi variables (1xN)
    x: Any
    # Casadi parameters
    p: Any
    # Indices of the integer variables
    idx_x_integer: List[float]
    # Flag if the hessian is not positive semi-definite
    hessian_not_psd: bool = False

    # SOC-type constraints
    h: Optional[Any] = None

    # Guass-Newton Hessian (NXN)
    gn_hessian: Optional[Any] = None
    # Indices of the linear constraints
    idx_g_lin: Optional[List[int]] = None
    # Indices of the linear integer constraints
    idx_g_lin_bin: Optional[List[int]] = None
    # Indices of the convex constraints
    idx_g_conv: Optional[List[int]] = None
    # Indices of the constraints that do not belong to any other category
    idx_g_other: Optional[List[int]] = None
    # Indices of dwelltime constraints
    idx_g_dwelltime: Optional[List[float]] = None
    idx_g_without_dwelltime: Optional[List[float]] = None

    # A precompiled NLP of the problem formulation
    precompiled_nlp: Optional[str] = None
    # Quadratic form or approximation of the objective function
    f_qp: Optional[Any] = None

    # Meta data of the problem
    meta: MetaData = field(default_factory=MetaData)

    def __post_init__(self):
        # Detect if integer vars are denoted using the casadi convention
        if len(self.idx_x_integer) == self.x.shape[0]:
            unique_elm_of_idx_x_integer = [i for i in set(self.idx_x_integer)]
            if unique_elm_of_idx_x_integer == [1]:
                self.idx_x_integer = [i for i in range(len(self.idx_x_integer))]
            elif unique_elm_of_idx_x_integer == [0]:
                self.idx_x_integer = []
            else:
                self.idx_x_integer = [
                    i
                    for i in range(len(self.idx_x_integer))
                    if self.idx_x_integer[i] == 1
                ]



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Implementation of the CIA algorithm."""

import copy
import datetime
import casadi as ca
import numpy as np
from typing import Tuple
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.solvers import SolverClass, Stats, MinlpProblem, MinlpData
from camino.settings import Settings
from camino.utils import toc, logging
from camino.utils.conversion import to_0d

try:
    from pycombina import BinApprox, CombinaBnB
except Exception:

    class BinApprox:
        pass

    class CombinaBnB:
        pass


logger = logging.getLogger(__name__)


def simulate(x0, u, f_dyn):
    N = u.shape[0]
    x = []
    for t in range(N):
        if t == 0:
            x.append(to_0d(f_dyn(x0, u[t, :])))
        else:
            x.append(to_0d(f_dyn(x[-1], u[t, :])))
    return np.array(x).flatten().tolist()


def to_list(dt, min_time, nr_b):
    """Create a min up or downtime list."""
    if isinstance(min_time, int):
        return [dt * min_time for _ in range(nr_b)]
    else:
        return [dt * min_time[i] for i in range(nr_b)]


class PycombinaSolver(SolverClass):
    """
    Create NLP solver.

    This solver solves an NLP problem. This is either relaxed or
    the binaries are fixed.
    """

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        """Create Pycombina solver."""
        super(PycombinaSolver, self).__init__(problem, stats, s)
        self.idx_x_integer = problem.idx_x_integer
        self.meta = copy.deepcopy(problem.meta)

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve CIA problem."""
        b_rel = to_0d(nlpdata.x_sol)[self.meta.idx_bin_control]
        if len(b_rel.shape) == 1:  # flatten array
            b_rel = b_rel.reshape(-1, self.meta.n_discrete_control)
        b_rel = np.hstack(
            [np.asarray(b_rel), np.array(1 - b_rel.sum(axis=1).reshape(-1, 1))]
        )  # Make sos1 structure

        # Ensure values are not out of range due to numerical effects
        b_rel[b_rel < 0] = 0
        b_rel[b_rel > 1.0] = 1

        N = b_rel.shape[0] + 1
        if isinstance(self.meta.dt, list):
            if isinstance(self.meta.dt[0], datetime.timedelta):
                time_array = [0]
                for t in self.meta.dt:
                    time_array.append(time_array[-1] + t.total_seconds())
                time_array = np.array(time_array)
            else:
                raise NotImplementedError()
        else:
            # Assumes uniform grid
            time_array = np.arange(0, N * self.meta.dt, self.meta.dt)
        binapprox = BinApprox(time_array, b_rel)

        value_set = False
        if self.meta.min_downtime is not None:
            value_set = True
            if isinstance(self.meta.min_downtime, np.ndarray):
                if self.meta.min_downtime.shape[0] == self.meta.n_discrete_control:
                    min_downtimes = np.concatenate(
                        [self.meta.min_downtime, np.zeros(1)]
                    )
                    binapprox.set_min_down_times(min_downtimes)
            else:
                binapprox.set_min_down_times(
                    to_list(self.meta.dt, self.meta.min_downtime, b_rel.shape[1])
                )

        if self.meta.min_uptime is not None:
            value_set = True
            if isinstance(self.meta.min_uptime, np.ndarray):
                if self.meta.min_uptime.shape[0] == self.meta.n_discrete_control:
                    min_uptimes = np.concatenate([self.meta.min_uptime, np.zeros(1)])
                    binapprox.set_min_up_times(min_uptimes)
            else:
                binapprox.set_min_up_times(
                    to_list(self.meta.dt, self.meta.min_uptime, b_rel.shape[1])
                )

        # binapprox.set_n_max_switches(...)
        # binapprox.set_max_up_times(...)

        if not value_set:
            raise Exception("Minimum uptime or downtime needs to be set!")

        combina = CombinaBnB(binapprox)
        combina.solve()
        b_bin = binapprox.b_bin[:-1, :].T.flatten()
        idx_bin_control = np.array(self.meta.idx_bin_control).flatten().tolist()
        nlpdata.x_sol[idx_bin_control] = b_bin

        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import casadi as ca
from camino.settings import GlobalSettings, Settings
from camino.problems import MinlpProblem, MinlpData, MetaDataOcp
from camino.problems.dsc import Description
from camino.utils.integrators import integrate_ie


def particle_trajectory():
    N = 100
    t_f = 10
    dt = t_f / N
    N_uptime = int(0.5 // dt)
    t = GlobalSettings.CASADI_VAR.sym("t")
    r = ca.Function("r", [t], [ca.sin(t) * 0.5 + 1])

    x = GlobalSettings.CASADI_VAR.sym("x")
    v = GlobalSettings.CASADI_VAR.sym("v")
    X = ca.vertcat(x, v)
    X_lb = [-100, -100]
    X_ub = [100, 100]
    X_0 = [0, 0]
    X_discrete = [0, 0]

    # Controls
    U = GlobalSettings.CASADI_VAR.sym("u", 3)
    U_lb = [-1, 0, 0]
    U_ub = [1, 1, 1]
    U_discrete = [1, 0, 0]

    # Dynamics
    x_dot = ca.vertcat(
        v,
        U[0],
    )

    # Terminal cost
    stage_cost = ca.Function("stage_cost", [X, U, t], [(x - r(t)) ** 2])

    # Discretize
    F = integrate_ie(X, U, x_dot, dt)

    # Create OCP
    dsc = Description()
    xp = dsc.add_parameters("x0", X.numel(), X_0)
    Up = 0
    u_vector = []
    for i in range(N):
        xk = dsc.sym("x", X.numel(), X_lb, X_ub, X_0, X_discrete)
        uk = dsc.sym("u", U.numel(), U_lb, U_ub, discrete=U_discrete)
        dsc.eq(Up, uk[0] - uk[1] + uk[2])
        u_vector.append(uk[1] + uk[2])
        dsc.eq(F(xp, uk, xk), 0)
        dsc.f += dt * stage_cost(xk, uk, (i + 1) * dt)
        xp, Up = xk, uk[0]

    for i in range(N - N_uptime):
        dsc.leq(sum(u_vector[i : i + N_uptime + 1]), 1)

    problem = dsc.get_problem()
    problem.meta = MetaDataOcp(
        dt=1,
        initial_state=X_0,
        n_state=X.numel(),
        n_continuous_control=1,
        n_discrete_control=0,
        idx_state=np.vstack(dsc.get_indices("x")),
        idx_control=np.vstack(dsc.get_indices("u"))[:, 0],
    )
    data = dsc.get_data()
    s = Settings()
    s.MIP_SETTINGS_ALL["gurobi"].update(
        {
            "gurobi.PoolSearchMode": 1,
            "gurobi.PoolSolutions": 3,
        }
    )
    return problem, data, s



"""
Simulator.

Adrian Buerger, 2022
Adapted by Wim Van Roy and Andrea Ghezzi, 2023
"""

import numpy as np
from datetime import timedelta

import logging
from camino.problems.solarsys.system import System, ca
from camino.problems.solarsys.ambient import Ambient, Timing
from camino.utils.cache import CachedFunction
from camino.utils.conversion import convert_to_flat_list


logger = logging.getLogger(__name__)


class Simulator(System):

    @property
    def time_grid(self):
        return self._timing.time_grid

    @property
    def x_data(self):
        try:
            return ca.DM(self._x_data)
        except AttributeError:
            msg = "Simulation results (states) not available yet, call solve() first."
            logging.error(msg)
            raise RuntimeError(msg)

    @property
    def u_data(self):
        try:
            return self._u_data
        except AttributeError:
            msg = "Continuous controls not available yet, call solve() first."
            logging.error(msg)
            raise RuntimeError(msg)

    @property
    def b_data(self):
        try:
            return self._b_data
        except AttributeError:
            msg = "Binary controls not available yet, call solve() first."
            logging.error(msg)
            raise RuntimeError(msg)

    @property
    def c_data(self):
        try:
            return self._c_data
        except AttributeError:
            msg = "Ambient parameters not available yet, call solve() first."
            logging.error(msg)
            raise RuntimeError(msg)

    def _get_ambient_paramaters(self):

        self._c_data = [
            convert_to_flat_list(
                self.nc, self.c_index, self._ambient.interpolate(self._ambient.get_t0())
            )
        ]
        tk = timedelta(seconds=0)
        for i in range(self._N - 1):
            tk += self._ambient.time_steps[i]
            self._c_data.append(
                convert_to_flat_list(
                    self.nc,
                    self.c_index,
                    self._ambient.interpolate(self._ambient.get_t0() + tk),
                )
            )
        self._c_data = ca.DM(self._c_data)

    def __init__(self, ambient: Ambient, N: int):

        logger.debug("Initializing simulator ...")
        super().__init__()
        self._ambient = ambient
        self._N = N
        self._integrator = self.get_integrator()
        self._get_ambient_paramaters()
        logger.debug("Simulator initialized.")

    def _set_initial_state(self):

        self._x_data = [
            convert_to_flat_list(
                self.nx, self.x_index, self.get_default_initial_state()
            )
        ]

    def _setup_controls(self):

        self._u_data = []
        self._b_data = []
        self._remaining_min_up_time = 0

    def _initialize_controls(self):

        self._u_data.append(np.zeros(self.nu))
        self._b_data.append(np.zeros(self.nb))

    def _set_b_ac(self, pos):

        T_hts = self._x_data[-1][self.x_index["T_hts"][0]]
        T_lts = self._x_data[-1][self.x_index["T_lts"]]

        T_amb = self.c_data[pos, self.c_index["T_amb"]]

        if self._remaining_min_up_time > 0:

            self._b_data[-1][self.b_index["b_ac"]] = self._b_data[-2][
                self.b_index["b_ac"]
            ]

            if self._b_data[-1][self.b_index["b_ac"]] == 1:
                self._remaining_min_up_time -= self._ambient.time_steps[
                    pos
                ].total_seconds()

        elif (T_hts < self.p["T_ac_ht_min"]) or (T_lts < self.p["T_ac_lt_min"]):

            self._b_data[-1][self.b_index["b_ac"]] = 0

        elif (
            (T_hts > self.p["T_ac_ht_min"] + self.p_csim["dT_ac_ht"])
            or (T_lts < self.p["T_ac_lt_max"] - self.p_csim["dT_ac_lt"])
            or (T_lts > self.p["T_ac_lt_min"] + self.p_csim["dT_ac_lt"])
        ):

            self._b_data[-1][self.b_index["b_ac"]] = 1

            try:
                if (self._b_data[-2][self.b_index["b_ac"]] != 1) and (
                    self._b_data[-1][self.b_index["b_ac"]] == 1
                ):
                    self._remaining_min_up_time = (
                        self.p_op["acm"]["min_up_time"][self.b_index["b_ac"]]
                        - self._ambient.time_steps[pos].total_seconds()
                    )
            except IndexError:
                self._remaining_min_up_time = 0

    def _set_b_hp(self):

        T_lts = self._x_data[-1][self.x_index["T_lts"]]

        if self._b_data[-1][self.b_index["b_ac"]] == 1:

            self._b_data[-1][self.b_index["b_hp"]] = 0

        elif T_lts > 20.0:

            self._b_data[-1][self.b_index["b_hp"]] = 1

    def _set_v_ppsc(self):

        T_fpsc = self._x_data[-1][self.x_index["T_fpsc"]]
        T_vtsc = self._x_data[-1][self.x_index["T_vtsc"]]
        T_hts = self._x_data[-1][self.x_index["T_hts"][0]]

        dT = max(T_fpsc, T_vtsc) - T_hts

        v_ppsc = (
            self.p_op["v_ppsc"]["max"]
            / (self.p_csim["dT_sc_ub"] - self.p_csim["dT_sc_lb"])
        ) * (
            max(self.p_csim["dT_sc_lb"], min(self.p_csim["dT_sc_ub"], dT))
            - self.p_csim["dT_sc_lb"]
        )

        self._u_data[-1][self.u_index["v_ppsc"]] = v_ppsc

    def _set_p_mpsc(self):

        T_fpsc = self._x_data[-1][self.x_index["T_fpsc"]]
        T_vtsc = self._x_data[-1][self.x_index["T_vtsc"]]

        dT = T_fpsc - T_vtsc

        p_mpsc = (
            (self.p_op["p_mpsc"]["max"] - self.p_op["p_mpsc"]["min_real"])
            / (self.p_csim["dT_vtsc_fpsc_ub"] - self.p_csim["dT_vtsc_fpsc_lb"])
        ) * (
            max(self.p_csim["dT_vtsc_fpsc_lb"], min(self.p_csim["dT_vtsc_fpsc_ub"], dT))
            - self.p_csim["dT_vtsc_fpsc_lb"]
        )

        self._u_data[-1][self.u_index["p_mpsc"]] = p_mpsc

    def _set_v_pssc(self):

        T_shx_ssc = self._x_data[-1][self.x_index["T_shx_ssc"][-1]]
        T_hts = self._x_data[-1][self.x_index["T_hts"][0]]

        dT = T_shx_ssc - T_hts

        v_pssc = (
            self.p_op["v_pssc"]["max"]
            / (self.p_csim["dT_sc_ub"] - self.p_csim["dT_sc_lb"])
        ) * (
            max(self.p_csim["dT_sc_lb"], min(self.p_csim["dT_sc_ub"], dT))
            - self.p_csim["dT_sc_lb"]
        )

        self._u_data[-1][self.u_index["v_pssc"]] = v_pssc

    def _set_mdot_o_hts_b(self):

        T_shx_ssc = self._x_data[-1][self.x_index["T_shx_ssc"][-1]]
        T_sc_feed_max = self.p_op["T_sc"]["T_feed_max"]

        dT = T_shx_ssc - T_sc_feed_max

        mdot_o_hts_b = (
            (1.0 / (self.p_csim["dT_o_hts_b_ub"] - self.p_csim["dT_o_hts_b_lb"]))
            * (
                max(self.p_csim["dT_o_hts_b_lb"], min(self.p_csim["dT_o_hts_b_ub"], dT))
                - self.p_csim["dT_o_hts_b_lb"]
            )
            * self.p["mdot_ssc_max"]
            * self._u_data[-1][self.u_index["v_pssc"]]
        )

        self._u_data[-1][self.u_index["mdot_o_hts_b"]] = mdot_o_hts_b

    def _set_mdot_i_hts_b(self):

        T_hts_m = self._x_data[-1][self.x_index["T_hts"][1]]
        T_i_hts_b_active = self.p_csim["T_i_hts_b_active"]

        dT = T_hts_m - T_i_hts_b_active

        mdot_i_hts_b = (
            (1.0 / (self.p_csim["dT_i_hts_b_ub"] - self.p_csim["dT_i_hts_b_lb"]))
            * (
                max(self.p_csim["dT_i_hts_b_lb"], min(self.p_csim["dT_i_hts_b_ub"], dT))
                - self.p_csim["dT_i_hts_b_lb"]
            )
            * self._b_data[-1][self.b_index["b_ac"]]
            * self.p["mdot_ac_ht"]
        )

        self._u_data[-1][self.u_index["mdot_i_hts_b"]] = mdot_i_hts_b

    def _set_controls(self, pos):

        self._initialize_controls()

        self._set_b_ac(pos)
        self._set_b_hp()

        self._set_v_ppsc()
        self._set_p_mpsc()
        self._set_v_pssc()
        self._set_mdot_o_hts_b()
        self._set_mdot_i_hts_b()

    def _run_step(self, pos, step):
        try:
            self._x_data.append(
                np.squeeze(
                    self._integrator(
                        x0=self.x_data[-1, :],
                        p=ca.veccat(
                            step.total_seconds(),
                            self.c_data[pos, :],
                            self.u_data[-1],
                            self.b_data[-1],
                        ),
                    )["xf"]
                )
            )
        except RuntimeError:

            print(
                (
                    "An error occurred during integration, "
                    "repeating states of previous integration step"
                )
            )

    def _finalize_simulation_results(self):

        # self._x_data = ca.horzcat(*self.x_data).T
        self._u_data = ca.horzcat(*self.u_data).T
        self._b_data = ca.horzcat(*self.b_data).T

    def _run_simulation(self):

        logger.info("Running simulation ...")

        self._set_initial_state()
        self._setup_controls()

        for _, (pos, step) in zip(range(self._N), enumerate(self._ambient.time_steps)):
            self._set_controls(pos)
            self._run_step(pos, step)

        self._finalize_simulation_results()

        logger.info("Simulation finished.")

    def solve(self):

        self._run_simulation()

    def predict(self, n_steps=1):
        """
        Predict the next initial state.
        Starting always from first element of c_data, u_data, b_data
        """

        x_hat = self.x_data[0, :]

        for k in range(n_steps):
            k += 1  # TODO make initial time instant editable
            x_hat = self._integrator(
                x0=x_hat,
                p=ca.veccat(
                    self._ambient.time_steps[k].total_seconds(),
                    self.c_data[k, :],
                    self.u_data[k, :],
                    self.b_data[k, :],
                ),
            )["xf"]

        return x_hat


if __name__ == "__main__":
    from datetime import timedelta
    from ambient import Ambient
    from camino.utils import setup_logger, logging
    import matplotlib.pyplot as plt

    setup_logger(logging.DEBUG)

    timing = Timing()
    ambient = Ambient(timing)
    simulator = Simulator(ambient=ambient, N=timing.N)
    simulator.solve()
    # x_hat = simulator.predict()

    plt.figure()
    plt.plot(range(simulator.c_data.shape[0]), simulator.c_data[:, 0])
    plt.plot(range(simulator.c_data.shape[0]), simulator.c_data[:, 1])
    plt.plot(range(simulator.c_data.shape[0]), simulator.c_data[:, 2])
    plt.plot(range(simulator.c_data.shape[0]), simulator.c_data[:, 3])
    plt.plot(range(simulator.c_data.shape[0]), simulator.c_data[:, 4])
    plt.plot(range(simulator.c_data.shape[0]), simulator.c_data[:, 5])
    plt.yscale("log")

    plt.figure()
    plt.plot(range(simulator.x_data.shape[0]), simulator.x_data[:, 0])
    plt.plot(range(simulator.x_data.shape[0]), simulator.x_data[:, 4])
    plt.show()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Models with pipes."""

from typing import Union
from camino.settings import GlobalSettings
from camino.problems import MinlpProblem, MinlpData, MetaDataOcp
from camino.problems.dsc import Description
from camino.utils.integrators import integrate_rk4
from camino.solvers import inspect_problem, set_constraint_types
from camino.utils.cache import cache_data
import numpy as np
import casadi as ca


def create_double_tank_problem2(
    p_val=[2, 2.5], single_shooting=False
) -> Union[MinlpProblem, MinlpData]:
    """
    Implement the double tank problem.

    Taken from Abbasi et al. ECC 23, reimplemented to achieve nice sparsity pattern.

    """
    eps = 1e-3
    N = 300
    dt = 1 / 30
    T = N * dt
    min_uptime = int(0.5 / dt)
    alpha = 100
    beta = np.array([1.0, 1.2])
    gamma = 10
    demand = np.array([2 + 0.5 * np.sin(x) for x in np.linspace(0, T, N + 1)])

    nx = 2
    nu = 2
    x = GlobalSettings.CASADI_VAR.sym("x", nx)  # state
    u = GlobalSettings.CASADI_VAR.sym("u", nu)  # control
    x1dot = gamma * u[0] + u[1] - ca.sqrt(x[0] + eps)
    x2dot = ca.sqrt(x[0] + eps) - ca.sqrt(x[1] + eps)
    xdot = ca.vertcat(*[x1dot, x2dot])
    # F = integrate_ee(x, u, xdot, dt, m_steps=1)
    F = integrate_rk4(x, u, xdot, dt, m_steps=1)

    dsc = Description()
    Xk = dsc.add_parameters("Xk0", nx, p_val)
    BigM = 1e3
    Uprev = None
    for k in range(N):
        Uk = dsc.sym("Uk", nu, lb=[0, 0], ub=[1, gamma], w0=[0, 0], discrete=[1, 0])
        if Uprev is not None and min_uptime > 0:
            # Implementation details: you can switch OFF only there are
            # no switch UP happened in the last "min_uptime -1" steps
            Suk = dsc.sym(
                "Suk", 1, lb=0, ub=1, w0=0, discrete=True
            )  # switching up variable
            dsc.eq(Uk[0] - Uprev[0] - Suk, [0])  # switch up detection

            # retrieve the indexes of Suk for window [k-min_uptime+1, k]
            slicing_idx = np.array(dsc.get_indices("Suk"))[-min(min_uptime, k) + 1 :]
            # min-uptime constraint: sum of switch UP in the window must be \leq than Uk
            dsc.leq(ca.sum1(ca.vcat(*[dsc.w])[slicing_idx]) - Uk[0], [0])
            # --> Uk can be 0 only if the min_uptime window there are no switch UP actions.

        # Integrate till the end of the interval
        Xk_end = F(Xk, Uk)
        dsc.f += dt * alpha * (Xk[1] - demand[k]) ** 2
        dsc.f += dt * (beta[0] * gamma * Uk[0] + beta[1] * Uk[1])

        if single_shooting:
            Xk = Xk_end
        else:
            # New NLP variable for state at end of interval
            Xk = dsc.sym("Xk", nx, lb=0, ub=BigM, w0=0.5)
            dsc.eq(Xk_end, Xk)

        Uprev = Uk

    problem = dsc.get_problem()
    meta = MetaDataOcp(
        dt=dt,
        n_state=nx,
        n_continuous_control=nu - 1,
        n_discrete_control=1,
        initial_state=p_val,
        idx_control=np.vstack(dsc.get_indices("Uk"))[:, 1],
        idx_state=np.hstack(dsc.get_indices("Xk")),
        idx_bin_control=np.vstack(dsc.get_indices("Uk"))[:, 0],
        scaling_coeff_control=[gamma, 1],
        min_uptime=min_uptime,
    )
    problem.meta = meta
    data = dsc.get_data()

    set_constraint_types(
        problem, *cache_data(f"doubletank{N}", inspect_problem, problem, data)
    )

    return problem, data



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

import casadi as ca
import numpy as np
from camino.solvers import MinlpProblem, MinlpData
from camino.settings import Settings
from camino.settings import GlobalSettings

CALLBACK_INPUTS = dict()
for i, label in enumerate(ca.nlpsol_out()):
    CALLBACK_INPUTS[label] = i


class CheckNoDuplicate:
    """Check duplicates."""

    def __init__(self, problem: MinlpProblem, s: Settings):
        """Check if no duplicates pass through."""
        self.idx_x_integer = problem.idx_x_integer
        self.s = s
        self.old = []
        self.count = 0

    def __call__(self, nlpdata: MinlpData):
        """Check if no old solutions pass through."""
        if nlpdata.prev_solutions is None:
            return
        for sol in nlpdata.prev_solutions:
            new = sol["x"][self.idx_x_integer]
            for el in self.old:
                if np.allclose(el, new, equal_nan=False, atol=self.s.EPS):
                    print("Duplicate!")
                    self.count += 1
            self.old.append(new)

        if self.count > 10:
            raise Exception()


class DebugCallBack(ca.Callback):
    """
    Create a debug callback.

    Usage:
        options = {}
        mycallback = DebugCallBack('Name', nx, ng, np, options)
        mycallback.add_to_solver_opts(options)
        ... Construct your solver as usual with options 'options' ...

        Every few iterations, the values for x will be written to a file in
        the datafolder.

    :param name: name of the callback function
    :param nx: Nr of x variables
    :param ng: Nr of g constraints
    :param np: Nr of parameters p
    :param opts: Additional options
    """

    def __init__(self, name, nx: int, ng: int, np: int, opts=None):
        """Create the debug call back for casadi."""
        ca.Callback.__init__(self)
        if opts is None:
            opts = {}

        self.nx = nx
        self.ng = ng
        self.np = np
        self.iter_nr = 0
        self.name = name
        # Initialize internal objects
        self.construct(name, opts)

    def add_to_solver_opts(self, options, iter_per_callback=50):
        """
        Add to the solver options.

        :param options: options of the nlp
        :param iter_per_callback: nr of iters per callback
        """
        options["iteration_callback"] = self
        options["iteration_callback_step"] = iter_per_callback
        return options

    def get_n_in(self):
        """Get number of inputs."""
        return ca.nlpsol_n_out()

    def get_n_out(self):
        """Get number of outputs."""
        return 1

    def get_sparsity_in(self, i):
        """Get sparsity of matrices."""
        n = ca.nlpsol_out(i)
        if n == "f":
            return ca.Sparsity.scalar()
        elif n in ("x", "lam_x"):
            return ca.Sparsity.dense(self.nx)
        elif n in ("g", "lam_g"):
            return ca.Sparsity.dense(self.ng)
        elif n in ("p", "lam_p"):
            return ca.Sparsity.dense(self.np)
        else:
            return ca.Sparsity(0, 0)

    def save(self, x):
        """Save the x variable."""
        self.iter_nr += 1
        np.save(GlobalSettings.DATA_FOLDER + f"/x_{self.name}_{self.iter}", x.full())

    def eval(self, arg):
        """Evaluate the callback."""
        x = arg[CALLBACK_INPUTS["x"]]
        self.save(x)
        return [0]



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Test if all combinations work."""

import unittest
from parameterized import parameterized
from camino.problems.problem_collection import PROBLEMS
from camino.solver import SOLVER_MODES
from camino.settings import Settings
from camino.runner import runner
import timeout_decorator

options = (
    [("cia", "doubletank2")]
    + [
        (solver, "dummy")
        for solver in SOLVER_MODES.keys()
        if solver
        not in ("cia", "s-tr-milp", "ampl", "nlp", "mip")  # Almost all solvers
    ]
    + [
        (solver, problem)
        for solver in [
            "s-b-miqp",
        ]
        for problem in PROBLEMS.keys()
        if problem
        not in [
            # Exclude duplicates
            "dummy2",
            # Exclude difficult problems with long runtimes
            "alan",
            "stcs",
            "to_car",
            "doubletank2",
            "doubletank",
            "stcs-simplified",
            # Exclude some errors:
            "unstable_ocp",
            "particle",
            "from_nlpsol_dsc",
            "sign_check",
            # Interfaces:
            "nl_file",
            "from_sto",
            "nosnoc",
            "from_nlpsol_dsc",
        ]
    ]
)

obj_val = {
    "dummy": 8.41,
    "doublepipe": 1,
    "dummy2": -3,
    "doubletank": 18.6826,
    "sign_check": 9,
    "nonconvex": 0.0567471,
    "gearbox": 6550.833,
    "particle": 1.3797,
    "unstable_ocp": -0.05129210,
    "gearbox_complx": 233.705,
    "gearbox_int": 14408.4375,
}

# Number of digits after the comma to be accurate:
obj_tolerance_default = 3
obj_tolerance = {
    "dummy": 2,
    "doubletank": 1,
    "nonconvex": 0,
    "particle": 2,
}
obj_tolerance_heuristic = {
    "nlp": -3,
    # Heuristics
    "rofp": -3,
    "ofp": -3,
    "fp": -3,
    # Disable:
    "cia": -100,
    "bonmin-qg": -100,
    "bonmin-hyb": -100,
    "nlp-fxd": -100,
}
sol_not_valid = ["nlp", "cia"]


class TestSolver(unittest.TestCase):
    """Test."""

    @parameterized.expand(options)
    @timeout_decorator.timeout(20.0)
    def test_solver(self, mode, problem_name):
        """Test runner."""
        s = Settings(TIME_LIMIT=10.0)
        s.MIP_SETTINGS_ALL["gurobi"]["gurobi.TimeLimit"] = 5.0
        stats, data = runner(mode, problem_name, None, None)
        desired_obj = obj_val.get(problem_name, -1)
        desired_tol = obj_tolerance.get(problem_name, obj_tolerance_default)
        desired_tol += obj_tolerance_heuristic.get(mode, 0)
        if desired_tol > -10:
            self.assertAlmostEqual(
                data.obj_val / desired_obj,
                1,
                desired_tol,
                msg=f"Failed for {mode} & {problem_name}",
            )


if __name__ == "__main__":
    unittest.main()
